/*
 * Copyright 1997-2003 Samuel Audet <guardia@step.polymtl.ca>
 *                     Taneli Lepp <rosmo@sektori.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* Code to manage PM123's bookmarks */

#define INCL_PM
#define INCL_DOS
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>

#include <utilfct.h>

#include "format.h"
#include "decoder_plug.h"
#include "pm123.h"
#include "bookmark.h"

extern HWND container, hwndPlaylist;

typedef struct
{
   char *filename;
   char *desc;

} ENTRY;

typedef struct
{
   ENTRY *entries;
   int num_entries;
   int entries_size;
} BOOKMARKS;


BOOKMARKS mbk = {0}; // menu bookmark entries


BOOL read_bookmarks(char *f, BOOKMARKS *b)
{
   int i;
   FILE *playlist;
   char line[512];

   playlist = fopen(f, "r");
   if(playlist == NULL)
      return FALSE;

   i = 0;
   while(!feof(playlist))
   {
      fgets(line, sizeof(line), playlist);
      blank_strip(line);

      if(i >= b->entries_size)
      {
         int j;
         if(b->entries_size <= 0)
            b->entries_size = 32;
         else
            b->entries_size *= 2;
         b->entries = (ENTRY *) realloc(b->entries, b->entries_size*sizeof(*b->entries));
         for(j = i; j < b->entries_size; j++)
         {
            b->entries[j].filename = NULL;
            b->entries[j].desc = NULL;
         }
      }

      if(line[0] != 0 && line[0] != '#' && line[0] != '>' && line[0] != '<')
      {
         free(b->entries[i].filename);
         b->entries[i].filename = (char *) malloc(strlen(line)+1);
         strcpy(b->entries[i].filename,line);
      }
      // '<' entry is after the filename
      else if(line[0] == '<')
      {
         free(b->entries[i].desc);
         b->entries[i].desc = (char *) malloc(strlen(line));
         strcpy(b->entries[i].desc,line+1);

         i++;
      }
   }
   b->num_entries = i;
   fclose(playlist);
   return TRUE;
}

void free_entry(ENTRY *entry)
{
   free(entry->filename);
   free(entry->desc);
   entry->filename = NULL;
   entry->desc = NULL;
}

void free_bookmarks(BOOKMARKS *b)
{
   int i;
   for(i = 0; i < b->num_entries; i++)
   {
      free_entry(&b->entries[i]);
   }
   free(b->entries);
   b->entries = NULL;
   b->entries_size = 0;
   b->num_entries  = 0;
}


BOOL save_bookmarks(char *f, BOOKMARKS *b)
{
   int i;
   FILE *playlist;
   char bakfile[512];

   char drive[_MAX_DRIVE], path[_MAX_PATH], fname[_MAX_FNAME];
   _splitpath(f, drive, path, fname, NULL);
   sprintf(bakfile,"%s%s%s.bak",drive,path,fname);

   remove(bakfile);
   rename(f,bakfile);

   playlist = fopen(f, "w");
   if(playlist == NULL)
      return FALSE;

   fprintf(playlist,
      "#\n"
      "# Bookmark list created with %s\n"
      "# Do not modify! This file is compatible with the playlist format,\n"
      "# but information written in this file is different.\n"
      "#\n", VERSION);

   for(i = 0; i < b->num_entries; i++)
   {
      int rc = fprintf(playlist,"%s\n<%s\n",b->entries[i].filename,b->entries[i].desc);
      if(rc < 0)
      {
         fclose(playlist);
         return FALSE;
      }
   }

   fprintf(playlist, "# End of bookmark list\n");

   if(fclose(playlist) == 0)
   {
      remove(bakfile);
      return TRUE;
   }
   else
      return FALSE;
}

BOOL rebuild_bookmarks(HWND hListbox, BOOKMARKS *newbk)
{
   int i = 0;
   short count = LONGFROMMR(WinSendMsg(hListbox,LM_QUERYITEMCOUNT,0,0));
   newbk->entries = malloc(count * sizeof(ENTRY));

   for(i = 0; i < count; i++)
   {
      ENTRY *anEntry = PVOIDFROMMR(WinSendMsg(hListbox,LM_QUERYITEMHANDLE,MPFROMSHORT(i),0));
      newbk->entries[i].filename = (char*) strdup(anEntry->filename);
      newbk->entries[i].desc = (char*) strdup(anEntry->desc);
   }
   newbk->num_entries = newbk->entries_size = i;
   return TRUE;
}

void move_item(HWND hListbox, int i, int j)
{
   short index;
   ENTRY *anEntry = PVOIDFROMMR(WinSendMsg(hListbox,LM_QUERYITEMHANDLE,MPFROMSHORT(i),0));

   WinSendMsg(hListbox,LM_DELETEITEM,MPFROMSHORT(i),0);
   index = LONGFROMMR(WinSendMsg(hListbox,LM_INSERTITEM,MPFROMSHORT(j),MPFROMP(anEntry->desc)));
   WinSendMsg(hListbox,LM_SETITEMHANDLE,MPFROMSHORT(index),MPFROMP(anEntry));

   WinSendMsg(hListbox,LM_SELECTITEM,MPFROMSHORT(index),MPFROMLONG(TRUE));
}

PFNWP wpListbox = NULL;

MRESULT EXPENTRY wpBkListbox(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   switch(msg)
   {
      case WM_CONTEXTMENU:
      {
         HWND hwndParent = WinQueryWindow(hwnd,QW_PARENT);

         if (SHORT2FROMMP(mp2) == TRUE)
         WinPopupMenu(hwnd, hwndParent, WinLoadMenu(HWND_OBJECT, 0, BOOKMARK_POPMENU), 10, 10, 0,
                      PU_HCONSTRAIN | PU_VCONSTRAIN | PU_MOUSEBUTTON1 | PU_KEYBOARD);
         else
         WinPopupMenu(hwnd, hwndParent, WinLoadMenu(HWND_OBJECT, 0, BOOKMARK_POPMENU), SHORT1FROMMP(mp1), SHORT2FROMMP(mp1), 0,
                      PU_HCONSTRAIN | PU_VCONSTRAIN | PU_MOUSEBUTTON1 | PU_KEYBOARD);
         return 0;
      }

   }

   return wpListbox(hwnd, msg, mp1, mp2);
}

MRESULT EXPENTRY wpEditBk(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   ENTRY *anEntry = WinQueryWindowPtr(hwnd, 0);
   HWND hFilename = WinWindowFromID(hwnd,EF_FILENAME);
   HWND hDesc = WinWindowFromID(hwnd,EF_DESC);

   switch(msg)
   {
      case WM_INITDLG:
         anEntry = PVOIDFROMMP(mp2);
         WinSetWindowPtr(hwnd, 0, anEntry);
         WinSendMsg(hFilename,EM_SETTEXTLIMIT,MPFROMSHORT(512),0);
         WinSendMsg(hDesc,EM_SETTEXTLIMIT,MPFROMSHORT(512),0);
         WinSetWindowText(hFilename,anEntry->filename);
         WinSetWindowText(hDesc,anEntry->desc);
         WinSetFocus(HWND_DESKTOP,hDesc);
         return MRFROMLONG(TRUE);

      case WM_CLOSE:
      {
         int lFilenameSize = WinQueryWindowTextLength(hFilename)+1;
         int lDescSize = WinQueryWindowTextLength(hDesc)+1;
         free_entry(anEntry);
         anEntry->filename = malloc(lFilenameSize);
         anEntry->desc = malloc(lDescSize);
         WinQueryWindowText(hFilename,lFilenameSize,anEntry->filename);
         WinQueryWindowText(hDesc,lDescSize,anEntry->desc);

         WinDismissDlg(hwnd,TRUE);
         break;
      }

      case WM_COMMAND:
      {
         CMDMSG *cmdmsg = COMMANDMSG(&msg);
         switch(cmdmsg->cmd)
         {
            case DID_OK:
            {
               LONG lFilenameSize = WinQueryWindowTextLength(hFilename)+1;
               LONG lDescSize = WinQueryWindowTextLength(hDesc)+1;
               free_entry(anEntry);
               anEntry->filename = malloc(lFilenameSize);
               anEntry->desc = malloc(lDescSize);
               WinQueryWindowText(hFilename,lFilenameSize,anEntry->filename);
               WinQueryWindowText(hDesc,lDescSize,anEntry->desc);

               WinDismissDlg(hwnd,TRUE);
               break;
            }

            case DID_CANCEL:
               WinDismissDlg(hwnd,FALSE);
               break;
         }

      }

   }

   return WinDefDlgProc(hwnd, msg, mp1, mp2);
}

MRESULT EXPENTRY wpBookmark(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   HWND hListbox = WinWindowFromID(hwnd,LB_BOOKMARKS);

   switch(msg)
   {
      case WM_WINDOWPOSCHANGED:
      {
         SWP *swp = PVOIDFROMMP(mp1);

         if( ((swp->fl & SWP_SIZE) || (swp->fl & SWP_MOVE)) && WinIsWindowVisible(hwnd) )
            cfg.Bookmark = *swp;

         if( ((swp->fl & SWP_SIZE) && (swp->fl & SWP_SHOW)) ||
             ((swp->fl & SWP_SIZE) && WinIsWindowVisible(hwnd)) )
         {
            SWP swpbutton = {0};
            SWP swpmenu = {0};
            int ty, bx, by;

            RECTL client = {swp->x, swp->y, swp->x+swp->cx, swp->y+swp->cy};
            WinCalcFrameRect(hwnd,&client,TRUE);
            WinMapWindowPoints(HWND_DESKTOP,hwnd,(POINTL *) &client,2);

            WinQueryWindowPos(WinWindowFromID(hwnd,DID_OK),&swpbutton);
            WinQueryWindowPos(WinWindowFromID(hwnd,FID_MENU),&swpmenu);

            WinEnableWindowUpdate(hwnd,FALSE);
            WinDefDlgProc (hwnd, msg, mp1, mp2);

            WinSetWindowPos(WinWindowFromID(hwnd, LB_BOOKMARKS),
                            HWND_TOP,
                            client.xLeft + 2,
                            client.yBottom + swpbutton.cy + 6,
                            client.xRight - client.xLeft - 4,
                            client.yTop - client.yBottom - swpbutton.cy - 6 - 2,
                            SWP_ACTIVATE | SWP_MOVE | SWP_SIZE | SWP_SHOW | SWP_ZORDER);

            WinEnableWindowUpdate(hwnd,TRUE);

            return 0;
         }

         break;
      }

      case WM_INITDLG:
      {
         int i;
         BOOKMARKS *bk = malloc(sizeof(*bk));

         memset(bk,0,sizeof(*bk));
         read_bookmarks(BOOKMARK_FILENAME,bk);
         WinSetWindowPtr(hwnd, 0, bk);

         for(i = 0; i < bk->num_entries; i++)
         {
            short index = LONGFROMMR(WinSendMsg(hListbox,LM_INSERTITEM,MPFROMSHORT(LIT_END),MPFROMP(bk->entries[i].desc)));
            WinSendMsg(hListbox,LM_SETITEMHANDLE, MPFROMSHORT(index),MPFROMP(&bk->entries[i]));
         }

         WinLoadMenu(hwnd,NULLHANDLE,BOOKMARK_MENU);
         WinSendMsg(hwnd,WM_UPDATEFRAME,MPFROMLONG(FCF_MENU),0);
         do_warpsans(hwnd);

         wpListbox = WinSubclassWindow(hListbox,(PFNWP)wpBkListbox);

         WinSetWindowPos(hwnd,HWND_TOP,cfg.Bookmark.x,cfg.Bookmark.y,
                         cfg.Bookmark.cx, cfg.Bookmark.cy,
                         SWP_SIZE | SWP_SHOW | SWP_MOVE);
         break;
      }

      case WM_CLOSE:
      {
         BOOKMARKS *bk = WinQueryWindowPtr(hwnd,0);
         BOOKMARKS newbk = {0};
         if(rebuild_bookmarks(hListbox,&newbk))
         {
            save_bookmarks(BOOKMARK_FILENAME,&newbk);
            free_bookmarks(&newbk);
         }
         free_bookmarks(bk);
         break;
      }

      case WM_COMMAND:
      {
         CMDMSG *cmdmsg = COMMANDMSG(&msg);
         short selected = LONGFROMMR(WinSendMsg(hListbox,LM_QUERYSELECTION,MPFROMSHORT(LIT_FIRST),0));
         BOOKMARKS *bk = WinQueryWindowPtr(hwnd,0);
         ENTRY *anEntry = PVOIDFROMMR(WinSendMsg(hListbox,LM_QUERYITEMHANDLE,MPFROMSHORT(selected),0));
         short count = LONGFROMMR(WinSendMsg(hListbox,LM_QUERYITEMCOUNT,0,0));

         switch(cmdmsg->cmd)
         {
            case PB_UP:
               if(selected > 0 && anEntry != NULL)
                  move_item(hListbox,selected,selected-1);
               return 0;

            case PB_DOWN:
               if(selected < count-1 && anEntry != NULL)
                  move_item(hListbox,selected,selected+1);
               return 0;

            case IDM_BK_OPEN:
               if(anEntry != NULL)
                  load_singlefile(anEntry->filename, container, hwndPlaylist);
               return 0;

            case IDM_BK_ADDTOPL:
               if(anEntry != NULL)
                  playlist_post_file(anEntry->filename);
               return 0;

            case IDM_BK_EDIT:
               if(selected != LIT_NONE && anEntry != NULL)
               {
                  BOOL DidOK = WinDlgBox(HWND_DESKTOP, hwnd, wpEditBk, NULLHANDLE, EDITBK_DLG, anEntry);
                  if(DidOK)
                     WinSendMsg(hListbox,LM_SETITEMTEXT,MPFROMSHORT(selected),MPFROMP(anEntry->desc));
               }
               return 0;

            case IDM_BK_REMOVE:
               if(selected != LIT_NONE && anEntry != NULL)
               {
                  free_entry(anEntry);
                  WinSendMsg(hListbox,LM_DELETEITEM,MPFROMSHORT(selected),0);
               }
               return 0;

            case DID_OK:
            {
               BOOKMARKS newbk = {0};
               if(rebuild_bookmarks(hListbox,&newbk))
               {
                  save_bookmarks(BOOKMARK_FILENAME,&newbk);
                  free_bookmarks(&newbk);
               }
            }
            case DID_CANCEL:
               free_bookmarks(bk);
               break;
         }
         break;
      }

      case WM_CONTROL:
      {
         short selected = LONGFROMMR(WinSendMsg(hListbox,LM_QUERYSELECTION,MPFROMSHORT(LIT_FIRST),0));
         ENTRY *anEntry = PVOIDFROMMR(WinSendMsg(hListbox,LM_QUERYITEMHANDLE,MPFROMSHORT(selected),0));

         if(SHORT1FROMMP(mp1) == LB_BOOKMARKS)
         {
            switch(SHORT2FROMMP(mp1))
            {
               case LN_ENTER:
                  if(anEntry != NULL)
                     load_singlefile(anEntry->filename, container, hwndPlaylist);
                  break;
            }
         }
         break;
      }

   }
   return WinDefDlgProc(hwnd, msg, mp1, mp2);
}

BOOL process_possible_bookmark(USHORT cmd)
{
   int i = cmd-IDM_M_BOOKMARKS-1;
   if(i >= 0 && i < mbk.num_entries)
   {
      load_singlefile(mbk.entries[i].filename, container, hwndPlaylist);
      return TRUE;
   }
   else
      return FALSE;
}

void add_bookmark(char *filename, char *desc)
{
   BOOKMARKS bk = {0};
   read_bookmarks(BOOKMARK_FILENAME, &bk);

   // bk.entries is allocated if no bookmark file was loaded
   // or if there is not enough space
   if(bk.entries_size <= bk.num_entries)
   {
      bk.entries_size++;
      bk.entries = (ENTRY *) realloc(bk.entries, bk.entries_size*sizeof(*bk.entries));
   }
   bk.entries[bk.num_entries].filename = malloc(strlen(filename)+1);
   bk.entries[bk.num_entries].desc = malloc(strlen(desc)+1);

   strcpy(bk.entries[bk.num_entries].filename,filename);
   strcpy(bk.entries[bk.num_entries].desc,desc);

   bk.num_entries++;

   if(!save_bookmarks(BOOKMARK_FILENAME, &bk))
      WinMessageBox(HWND_DESKTOP, HWND_DESKTOP, "Error saving bookmark file.", "Error", 0, MB_ERROR | MB_OK);
}

void load_bookmark_menu(HWND hMenu)
{
   MENUITEM mi;
   int i = 0;
   int count;

   count = LONGFROMMR(WinSendMsg(hMenu, MM_QUERYITEMCOUNT, 0, 0));
   for(i = 0; i < count; i++)
   {
      short item = LONGFROMMR(WinSendMsg(hMenu, MM_ITEMIDFROMPOSITION, MPFROMSHORT(0), 0));
      WinSendMsg(hMenu,MM_DELETEITEM, MPFROM2SHORT(item, FALSE), 0);
   }

   free_bookmarks(&mbk);
   if(!read_bookmarks(BOOKMARK_FILENAME,&mbk) || mbk.num_entries == 0)
   {
      mi.iPosition = MIT_END;
      mi.afStyle = MIS_TEXT;
      mi.afAttribute = MIA_DISABLED;
      mi.id = IDM_M_BOOKMARKS + 1;
      mi.hwndSubMenu = (HWND)NULLHANDLE;
      mi.hItem = 0;
      WinSendMsg(hMenu, MM_INSERTITEM, MPFROMP(&mi), MPFROMP("No bookmarks"));
   }
   else
   {
      for(i = 0; i < mbk.num_entries; i++)
      {
         mi.iPosition = MIT_END;
         mi.afStyle = MIS_TEXT;
         mi.afAttribute = 0;
         mi.id = IDM_M_BOOKMARKS + i + 1;
         mi.hwndSubMenu = (HWND)NULLHANDLE;
         mi.hItem = 0;
         WinSendMsg(hMenu, MM_INSERTITEM, MPFROMP(&mi), MPFROMP(mbk.entries[i].desc));
      }
   }

}
