/*

gbmv.c - General Bitmap Viewer

*/

#define	INCL_DOS
#define	INCL_WIN
#define	INCL_GPI
#define	INCL_DEV
#define	INCL_BITMAPFILEFORMAT
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
//#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "gbm.h"
#include "gbmerr.h"
#include "gbmht.h"
#include "gbmv.h"

extern char startpath[256];

static VOID Error(HWND hwnd, CHAR *szFmt, ...)
	{
	}

static VOID Fatal(HWND hwnd, CHAR *sz)
	{
	Error(hwnd, sz);
	}

static VOID Usage(HWND hwnd)
	{
	Fatal(hwnd, "Usage: gbmv [-e] [-h] filename.ext{,opt}");
	}

static VOID Warning(HWND hwnd, CHAR *debugs)
	{
//     FILE *debug;
//     debug = fopen("skin.log", "at");
//     fprintf(debug, "%s\n", debugs);
//     fclose(debug);

	}

static HBITMAP hbmBmp;
static LONG lColorBg, lColorFg;

static BOOL LoadBitmap(HWND hwnd, CHAR *szFn, CHAR *szOpt, GBM *gbm, GBMRGB *gbmrgb, BYTE **ppbData)
{
	GBM_ERR rc;
	int fd, ft;
	ULONG cb;
	USHORT usStride;
    char save[256];
    char filename[256];

    Warning(hwnd, szFn);
    strcpy(filename, szFn);

    if ( gbm_guess_filetype(filename, &ft) != GBM_ERR_OK )
		{
        Warning(hwnd, "Can't deduce bitmap format from file extension");
        strcpy(save, filename);
        strcpy(filename, startpath);
        strcat(filename, save);
        if ( gbm_guess_filetype(filename, &ft) != GBM_ERR_OK ) { return ( FALSE ); }
        }

    if ( (fd = gbm_io_open(filename, O_RDONLY | O_BINARY)) == -1 )
		{
        strcpy(save, filename);
        strcpy(filename, startpath);
        strcat(filename, save);
        Warning(hwnd, "Can't open file");
        if ( (fd = gbm_io_open(filename, O_RDONLY | O_BINARY)) == -1 )
          return ( FALSE );
		}

    if ( (rc = gbm_read_header(filename, fd, ft, gbm, szOpt)) != GBM_ERR_OK )
		{
		gbm_io_close(fd);
        Warning(hwnd, "Can't read file header");
        return ( FALSE );
		}

	if ( (rc = gbm_read_palette(fd, ft, gbm, gbmrgb)) != GBM_ERR_OK )
		{
		gbm_io_close(fd);
        Warning(hwnd, "Can't read file palette");
        return ( FALSE );
		}

	usStride = ((gbm -> w * gbm -> bpp + 31)/32) * 4;
	cb = gbm -> h * usStride;
	if ( (*ppbData = malloc((int) cb)) == NULL )
		{
		gbm_io_close(fd);
        Warning(hwnd, "Out of memory");
        return ( FALSE );
		}

	if ( (rc = gbm_read_data(fd, ft, gbm, *ppbData)) != GBM_ERR_OK )
		{
		free(*ppbData);
		gbm_io_close(fd);
        Warning(hwnd, "Can't read file data");
        return ( FALSE );
		}

	gbm_io_close(fd);

	return ( TRUE );
	}

static BOOL MakeBitmap(
	HWND hwnd,
    GBM *gbm, GBMRGB *gbmrgb, BYTE *pbData,
	HBITMAP *phbm
	)
	{
	HAB hab = WinQueryAnchorBlock(hwnd);
	USHORT cRGB, usCol;
	SIZEL sizl;
	HDC hdc;
	HPS hps;
	struct
		{
		BITMAPINFOHEADER2 bmp2;
		RGB2 argb2Color [0x100];
		} bm;

	/* Got the data in memory, now make bitmap */

	memset(&bm, 0, sizeof(bm));

	bm.bmp2.cbFix     = sizeof(BITMAPINFOHEADER2);
	bm.bmp2.cx        = gbm -> w;
	bm.bmp2.cy        = gbm -> h;
	bm.bmp2.cBitCount = gbm -> bpp;
	bm.bmp2.cPlanes   = 1;

	cRGB = ( (1 << gbm -> bpp) & 0x1ff );
		/* 1 -> 2, 4 -> 16, 8 -> 256, 24 -> 0 */

	for ( usCol = 0; usCol < cRGB; usCol++ )
		{
		bm.argb2Color [usCol].bRed   = gbmrgb [usCol].r;
		bm.argb2Color [usCol].bGreen = gbmrgb [usCol].g;
		bm.argb2Color [usCol].bBlue  = gbmrgb [usCol].b;
		}

	if ( (hdc = DevOpenDC(hab, OD_MEMORY, "*", 0L, (PDEVOPENDATA) NULL, (HDC) NULL)) == (HDC) NULL )
		{
		Warning(hwnd, "DevOpenDC failure");
		return ( FALSE );
		}

	sizl.cx = bm.bmp2.cx;
	sizl.cy = bm.bmp2.cy;
	if ( (hps = GpiCreatePS(hab, hdc, &sizl, PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC)) == (HPS) NULL )
		{
		Warning(hwnd, "GpiCreatePS failure");
		DevCloseDC(hdc);
		return ( FALSE );
		}


	if ( cRGB == 2 )
{
static RGB2 argb2Black = { 0x00, 0x00, 0x00 };
static RGB2 argb2White = { 0xff, 0xff, 0xff };
bm.argb2Color [0] = argb2Black; /* Contrast */
bm.argb2Color [1] = argb2White; /* Reset */
}

	if ( (*phbm = GpiCreateBitmap(hps, &(bm.bmp2), CBM_INIT, (BYTE *) pbData, (BITMAPINFO2 *) &(bm.bmp2))) == (HBITMAP) NULL )
		{
		Warning(hwnd, "GpiCreateBitmap failure");
		GpiDestroyPS(hps);
		DevCloseDC(hdc);
		return ( FALSE );
		}

	GpiSetBitmap(hps, (HBITMAP) NULL);
	GpiDestroyPS(hps);
	DevCloseDC(hdc);

	return ( TRUE );
	}

HBITMAP gbm_load_bitmap(char *filename)
{
	QMSG qMsg;
    CHAR sz [255+1], szOpt[1];
	GBM gbm;
	GBMRGB gbmrgb [0x100];
	BYTE *pbData;
	BOOL fHalftone = FALSE, fErrdiff = FALSE, fOk;
	int i;
    static int initialized = 0;

//    if (!initialized)
//     {
      gbm_init();
//      initialized = 1;
//     }

    *szOpt = '\0';
    strcpy(sz, filename);

    if (LoadBitmap(HWND_DESKTOP, sz, szOpt, &gbm, gbmrgb, &pbData) )
      {
       HPS hps = WinGetPS(HWND_DESKTOP);
       HDC hdc = GpiQueryDevice(hps);
       fOk = MakeBitmap(HWND_DESKTOP, &gbm, gbmrgb, pbData, &hbmBmp);
       Warning(0, "Bitmap done");
       free(pbData);
       gbm_deinit();
       return (hbmBmp);
      }
     gbm_deinit();
     return (0);
}
