/*
 * Copyright 1997-2003 Samuel Audet <guardia@step.polymtl.ca>
 *                     Taneli Lepp <rosmo@sektori.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#define INCL_PM
#define INCL_DOS
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <utilfct.h>

#include "format.h"
#include "decoder_plug.h"
#include "output_plug.h"
#include "filter_plug.h"
#include "plugin.h"
#include "pm123.h"
#include "plugman.h"

void factory_settings(void)
{
   POINTL dsize[2] = {{13, 27}, {202, 138}};
   WinMapDlgPoints(HWND_DESKTOP,dsize,2,TRUE);

   memset(&cfg,0,sizeof(cfg));

   cfg.Bookmark.x = dsize[0].x;
   cfg.Bookmark.y = dsize[0].y;
   cfg.Bookmark.cx = dsize[1].x;
   cfg.Bookmark.cy = dsize[1].y;

   cfg.Properties.x = 50;
   cfg.Properties.y = 50;
   cfg.Properties.cx = 0; // not used...
   cfg.Properties.cy = 0;

   /* Anyway, the main window position */
   cfg.Main.x = 1;
   cfg.Main.y = 1;

   /* Playlist window size */
   cfg.Playlist.x = 1;
   cfg.Playlist.y = 115;
   cfg.Playlist.cx = 450;
   cfg.Playlist.cy = 400;

   /* Directory for MPEG files */
   strcpy(cfg.filedir, "");
   /* Directory for playlists  */
   strcpy(cfg.listdir, "");

   /* Default skin */
   *(cfg.defaultskin) = '\0';

   /* Default maximum volume, of course */
   cfg.defaultvol = 38;

   /* Default to DART */
  // cfg.soundsystem = 0;
   /* Default to stereo playing (no mixing) */
  // cfg.mono = 0;
   /* Don't play file automatically on load */
   cfg.playonload = 0;
   /* Don't auto use playlist on load */
   cfg.autouse = 0;

   /* Make the bolder font a default */
   cfg.font = 1;

   /* No 8-bit audio as default - 8 bit audio sucks! */
  // cfg.eightbit = 0;
   /* Downmixing sucks! */
  // cfg.downmix = 0;

   /* Device number 0 by default and shared by default */
  // cfg.deviceno = 0; cfg.shared = 1;

   /* 32 buffers by default */
  // cfg.buffers = 32;

   /* Flush buffers by default */
   cfg.trash = 1;

   /* Low priority by default */
   cfg.priority = 0;

   /* Bleah (tm) */
   cfg.shf = 0; cfg.rpt = 0; cfg.floatontop = 0;

   /* Poof (R) */
   cfg.playonuse = 0; cfg.scroll = 0;

   /* Last used stuph */
   memset(cfg.last[0], '\0', 256 * 10); memset(cfg.list[0], '\0', 256 * 10);

   cfg.first_run = time((time_t *)0);

   /* oOoo */
  // cfg.kludge48as44 = 0;  /* We really don't like this */
   cfg.bufsize      = 0;  /* 64kB rocks you */
   cfg.bufwait     = 1;   /* It's ok to fill the buffer first */
   strcpy(cfg.proxy, ""); /* No proxy */
   strcpy(cfg.auth, "");  /* No authentication */
   cfg.viewmode    = 0;   /* Filename owns you */

   /* Last but not the least: our signature (Seal of Quality :-)) */
   strcpy(cfg.id, "PM123 (c) 2000 Taneli Lepp, Samuel Audet");
   cfg.eof = (char)26;

   cfg.mode = 0;

   load_default_decoders();
   load_default_outputs();
   load_default_filters();

   cfg.show_playlist = 0;

   cfg.dock_windows = 1;
   cfg.dock_margin = 10;
}


void load_ini()
{
   HINI INIhandle;
   BUFSTREAM *b;
   int i;

   factory_settings();

   if((INIhandle = open_module_ini()) != NULLHANDLE)
   {
      load_ini_value(INIhandle,cfg.defaultvol);
      load_ini_value(INIhandle,cfg.playonload);
      load_ini_value(INIhandle,cfg.autouse);
//      load_ini_value(INIhandle,cfg.divider);
      load_ini_value(INIhandle,cfg.mode);
      load_ini_value(INIhandle,cfg.font);
//      load_ini_value(INIhandle,cfg.priority);
      load_ini_value(INIhandle,cfg.trash);
      load_ini_value(INIhandle,cfg.shf);
      load_ini_value(INIhandle,cfg.rpt);
      load_ini_value(INIhandle,cfg.floatontop);
      load_ini_value(INIhandle,cfg.playonuse);
      load_ini_value(INIhandle,cfg.scroll);
      load_ini_value(INIhandle,cfg.first_run);
      load_ini_value(INIhandle,cfg.viewmode);
      load_ini_value(INIhandle,cfg.bufwait);
      load_ini_value(INIhandle,cfg.bufsize);

      // window size and positions
      load_ini_value(INIhandle,cfg.Main);
      load_ini_value(INIhandle,cfg.Playlist);
      load_ini_value(INIhandle,cfg.Manager);
      load_ini_value(INIhandle,cfg.Equalizer);
      load_ini_value(INIhandle,cfg.Bookmark);
      load_ini_value(INIhandle,cfg.Properties);

      load_ini_string(INIhandle,cfg.filedir,sizeof(cfg.filedir));
      load_ini_string(INIhandle,cfg.listdir,sizeof(cfg.listdir));
      load_ini_string(INIhandle,cfg.savedir,sizeof(cfg.savedir));
      load_ini_string(INIhandle,cfg.cddrive,sizeof(cfg.savedir));
//      load_ini_string(INIhandle,cfg.pl_font);
//      load_ini_data(INIhandle,&cfg.pl_fg_font,sizeof(cfg.pl_fg_font));
//      load_ini_data(INIhandle,&cfg.pl_bg_font,sizeof(cfg.pl_bg_font));

      load_ini_string(INIhandle,cfg.proxy, sizeof(cfg.proxy));
      load_ini_string(INIhandle,cfg.auth, sizeof(cfg.auth));
      load_ini_string(INIhandle,cfg.defaultskin, sizeof(cfg.defaultskin));

      load_ini_value(INIhandle,cfg.last);
      load_ini_value(INIhandle,cfg.list);

      {
      void *decoders_list;
      ULONG size;

      load_ini_data_size(INIhandle, decoders_list, size);
      if(size > 0 && (decoders_list = malloc(size)) != NULL)
      {
         load_ini_data(INIhandle, decoders_list, size);
         b = open_bufstream(decoders_list, size);
         load_decoders(b);
         close_bufstream(b);
      }
      }

      {
      void *outputs_list;
      ULONG size;

      load_ini_data_size(INIhandle, outputs_list, size);
      if(size > 0 && (outputs_list = malloc(size)) != NULL)
      {
         load_ini_data(INIhandle, outputs_list, size);
         b = open_bufstream(outputs_list, size);
         load_outputs(b);
         close_bufstream(b);
      }
      }

      {
      void *filters_list;
      ULONG size;

      load_ini_data_size(INIhandle, filters_list, size);
      if(size > 0 && (filters_list = malloc(size)) != NULL)
      {
         load_ini_data(INIhandle, filters_list, size);
         b = open_bufstream(filters_list, size);
         load_filters(b);
         close_bufstream(b);
      }
      }

// -- loading equalizer

      for (i = 0; i < 20; i++) gains[i] = 1.0;
      for (i = 0; i < 20; i++) mutes[i] = 0;
      preamp = 1.0;

      load_ini_string(INIhandle,cfg.lasteq,sizeof(cfg.lasteq));
      amp_load_eq_file(cfg.lasteq, gains, mutes, &preamp);
      load_ini_value(INIhandle,cfg.eq_enabled);

      load_ini_value(INIhandle,cfg.show_playlist);

      load_ini_value(INIhandle,cfg.dock_margin);
      load_ini_value(INIhandle,cfg.dock_windows);

      close_ini(INIhandle);
   }

}

void save_ini()
{
   HINI INIhandle;
   BUFSTREAM *b;
   if((INIhandle = open_module_ini()) != NULLHANDLE)
   {
      save_ini_string(INIhandle,cfg.id);

      save_ini_value(INIhandle,cfg.defaultvol);
      save_ini_value(INIhandle,cfg.playonload);
      save_ini_value(INIhandle,cfg.autouse);
//      save_ini_value(INIhandle,cfg.divider);
      save_ini_value(INIhandle,cfg.mode);
      save_ini_value(INIhandle,cfg.font);
//      save_ini_value(INIhandle,cfg.priority);
      save_ini_value(INIhandle,cfg.trash);
      save_ini_value(INIhandle,cfg.shf);
      save_ini_value(INIhandle,cfg.rpt);
      save_ini_value(INIhandle,cfg.floatontop);
      save_ini_value(INIhandle,cfg.playonuse);
      save_ini_value(INIhandle,cfg.scroll);
      save_ini_value(INIhandle,cfg.first_run);
      save_ini_value(INIhandle,cfg.viewmode);
      save_ini_value(INIhandle,cfg.bufwait);
      save_ini_value(INIhandle,cfg.bufsize);

      // window size and positions
      save_ini_value(INIhandle,cfg.Main);
      save_ini_value(INIhandle,cfg.Playlist);
      save_ini_value(INIhandle,cfg.Manager);
      save_ini_value(INIhandle,cfg.Equalizer);
      save_ini_value(INIhandle,cfg.Bookmark);
      save_ini_value(INIhandle,cfg.Properties);

      save_ini_string(INIhandle,cfg.filedir);
      save_ini_string(INIhandle,cfg.listdir);
      save_ini_string(INIhandle,cfg.savedir);
      save_ini_string(INIhandle,cfg.cddrive);
//      save_ini_string(INIhandle,cfg.pl_font);
//      save_ini_data(INIhandle,&cfg.pl_fg_font,sizeof(cfg.pl_fg_font));
//      save_ini_data(INIhandle,&cfg.pl_bg_font,sizeof(cfg.pl_bg_font));

      save_ini_string(INIhandle,cfg.proxy);
      save_ini_string(INIhandle,cfg.auth);
      save_ini_string(INIhandle,cfg.defaultskin);

      save_ini_value(INIhandle,cfg.last);
      save_ini_value(INIhandle,cfg.list);

      {
      void *decoders_list;
      int size;

      b = create_bufstream(1024);
      save_decoders(b);
      size = get_buffer_bufstream(b, &decoders_list);
      save_ini_data(INIhandle,decoders_list,size);
      close_bufstream(b);
      }

      {
      void *outputs_list;
      int size;

      b = create_bufstream(1024);
      save_outputs(b);
      size = get_buffer_bufstream(b, &outputs_list);
      save_ini_data(INIhandle,outputs_list,size);
      close_bufstream(b);
      }

      {
      void *filters_list;
      int size;

      b = create_bufstream(1024);
      save_filters(b);
      size = get_buffer_bufstream(b, &filters_list);
      save_ini_data(INIhandle,filters_list,size);
      close_bufstream(b);
      }

      save_ini_string(INIhandle,cfg.lasteq);
      save_ini_value(INIhandle,cfg.eq_enabled);

      save_ini_value(INIhandle,cfg.show_playlist);

      save_ini_value(INIhandle,cfg.dock_windows);
      save_ini_value(INIhandle,cfg.dock_margin);

      close_ini(INIhandle);
   }
}
