/*
 * Copyright 1997-2003 Samuel Audet <guardia@step.polymtl.ca>
 *                     Taneli Lepp <rosmo@sektori.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* Code to manager new plug-in types (decoder, output and filter plug-ins). */
/* Visual plug-ins still use the old method in pm123.c and 123_util.c, but  */
/* should be ported to plugman.c with the rest.                             */

#define INCL_PM
#define INCL_DOS

#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <utilfct.h>

#include "format.h"
#include "output_plug.h"
#include "decoder_plug.h"
#include "filter_plug.h"
#include "plugin.h"
#include "plugman.h"
#include "pm123.h"

int out_set_active(int number);
ULONG check_plugin(HMODULE module, char *module_name, PLUGIN_QUERYPARAM *query_param);

BOOL add_decoder_plugin(DECODER *decoder);
BOOL add_output_plugin(OUTPUT *output);
BOOL add_filter_plugin(FILTER *filter);

void error_display(char *buf)
{
   WinMessageBox(HWND_DESKTOP, HWND_DESKTOP, buf, "Error", 0, MB_ERROR | MB_OK);
}


DECODER *decoders = NULL;
int num_decoders = 0;
int active_decoder = -1;
OUTPUT *outputs = NULL;
int num_outputs = 0;
int active_output = -1;
FILTER *filters = NULL;
int num_filters = 0;

BOOL load_module(char *module_name, HMODULE *module)
{
   ULONG rc;
   char error[128] = "";
   char temp[128] = "";

   rc = DosLoadModule(error, sizeof(error), module_name, module);

   if(rc != 0)
   {
      sprintf(temp,"Error: Could not load %s, %s, rc = %d.",module_name,error,rc);
      error_display(temp);
      return FALSE;
   }

   return TRUE;
}

void unload_module(HMODULE module)
{
   DosFreeModule(module);
}

BOOL load_function(HMODULE module, PFN *function, char *function_name, char *module_name)
{
   ULONG rc;
   char temp[128];
   rc = DosQueryProcAddr(module, 0L, function_name, function);

   if(rc != 0)
   {
      *function = NULL;
      sprintf(temp,"Error: Could not load %s from %s, rc = %d.",function_name, module_name, rc);
      error_display(temp);
      return FALSE;
   }

   return TRUE;

}

void load_decoder(DECODER *info)
{
   check_plugin(info->module, info->module_name, &info->query_param);

   load_function(info->module,(PFN*)&info->decoder_init,"decoder_init",info->module_name);
   load_function(info->module,(PFN*)&info->decoder_uninit,"decoder_uninit",info->module_name);
   load_function(info->module,(PFN*)&info->decoder_command,"decoder_command",info->module_name);
   load_function(info->module,(PFN*)&info->decoder_status,"decoder_status",info->module_name);
   load_function(info->module,(PFN*)&info->decoder_length,"decoder_length",info->module_name);
   load_function(info->module,(PFN*)&info->decoder_fileinfo,"decoder_fileinfo",info->module_name);
   load_function(info->module,(PFN*)&info->decoder_trackinfo,"decoder_trackinfo",info->module_name);
   load_function(info->module,(PFN*)&info->decoder_cdinfo,"decoder_cdinfo",info->module_name);

   load_function(info->module, (PFN*)&info->plugin_query, "plugin_query", info->module_name);
   if(info->query_param.configurable)
      load_function(info->module, (PFN*)&info->plugin_configure, "plugin_configure", info->module_name);

   info->w = NULL;

   info->enabled = TRUE;
}

void load_output(OUTPUT *info)
{
   check_plugin(info->module, info->module_name, &info->query_param);

   load_function(info->module,(PFN*)&info->output_init,"output_init",info->module_name);
   load_function(info->module,(PFN*)&info->output_uninit,"output_uninit",info->module_name);
   load_function(info->module,(PFN*)&info->output_command,"output_command",info->module_name);
   load_function(info->module,(PFN*)&info->output_play_samples,"output_play_samples",info->module_name);
   load_function(info->module,(PFN*)&info->output_playing_samples,"output_playing_samples",info->module_name);
   load_function(info->module,(PFN*)&info->output_playing_pos,"output_playing_pos",info->module_name);
   load_function(info->module,(PFN*)&info->output_playing_data,"output_playing_data",info->module_name);

   load_function(info->module, (PFN*)&info->plugin_query, "plugin_query", info->module_name);
   if(info->query_param.configurable)
      load_function(info->module, (PFN*)&info->plugin_configure, "plugin_configure", info->module_name);

   info->a = NULL;
}

void load_filter(FILTER *info)
{
   check_plugin(info->module, info->module_name, &info->query_param);

   load_function(info->module,(PFN*)&info->filter_init,"filter_init",info->module_name);
   load_function(info->module,(PFN*)&info->filter_uninit,"filter_uninit",info->module_name);
   load_function(info->module,(PFN*)&info->filter_play_samples,"filter_play_samples",info->module_name);

   load_function(info->module, (PFN*)&info->plugin_query, "plugin_query", info->module_name);
   if(info->query_param.configurable)
      load_function(info->module, (PFN*)&info->plugin_configure, "plugin_configure", info->module_name);

   info->f = NULL;

   info->enabled = TRUE;
}


void load_default_decoders()
{
   int i = 0;
   DECODER decoder;

   free(decoders);
   num_decoders = 0;

   strcpy(decoder.module_name, "mpg123");
   if(load_module(decoder.module_name, &decoder.module))
      add_decoder_plugin(&decoder);

   strcpy(decoder.module_name, "wavplay");
   if(load_module(decoder.module_name, &decoder.module))
      add_decoder_plugin(&decoder);

   strcpy(decoder.module_name, "cddaplay");
   if(load_module(decoder.module_name, &decoder.module))
      add_decoder_plugin(&decoder);
}

void load_default_outputs()
{
   int i = 0;
   OUTPUT output;

   free(outputs);
   num_outputs = 0;

   strcpy(output.module_name, "os2audio");
   if(load_module(output.module_name, &output.module))
      add_output_plugin(&output);

   strcpy(output.module_name, "wavout");
   if(load_module(output.module_name, &output.module))
      add_output_plugin(&output);

   out_set_active(0);
}

void load_default_filters()
{
   int i = 0;
   free(filters);
   num_filters = 1;
   filters = (FILTER *) malloc(num_filters * sizeof(FILTER));

   strcpy(filters[0].module_name, "realeq");
   if(load_module(filters[0].module_name, &filters[0].module))
   {
      i++;
      load_filter(filters);
   }
   else
      num_filters--;
}

void load_decoders(BUFSTREAM *b)
{
   int i, temp;
   int o = read_bufstream(b, &num_decoders, sizeof(int));
   if(o != sizeof(int))
   {
      load_default_decoders();
      return;
   }

   dec_set_active(-1);
   free(decoders);
   decoders = (DECODER *) malloc(num_decoders * sizeof(DECODER));

   for(i = 0; i < num_decoders; i++)
   {
      o = read_bufstream(b, &temp, sizeof(int));
      if(o != sizeof(int))
      {
         load_default_decoders();
         return;
      }
      o = read_bufstream(b, &decoders[i].module_name, temp);
      decoders[i].module_name[temp] = 0;
      if(o != temp)
      {
         load_default_decoders();
         return;
      }

      decoders[i].enabled = FALSE;
      if(load_module(decoders[i].module_name, &decoders[i].module))
      {
         load_decoder(&decoders[i]);

         o = read_bufstream(b, &temp, sizeof(int));
         if(o != sizeof(int))
            decoders[i].enabled = TRUE;
         else
            decoders[i].enabled = temp;
      }
   }
}


void load_outputs(BUFSTREAM *b)
{
   int i, temp;
   int o = read_bufstream(b, &num_outputs, sizeof(int));
   if(o != sizeof(int))
   {
      load_default_outputs();
      return;
   }

   out_set_active(-1);
   free(outputs);
   outputs = (OUTPUT *) malloc(num_outputs * sizeof(OUTPUT));

   for(i = 0; i < num_outputs; i++)
   {
      o = read_bufstream(b, &temp, sizeof(int));
      if(o != sizeof(int))
      {
         load_default_outputs();
         return;
      }
      o = read_bufstream(b, &outputs[i].module_name, temp);
      outputs[i].module_name[temp] = 0;
      if(o != temp)
      {
         load_default_outputs();
         return;
      }

      if(load_module(outputs[i].module_name, &outputs[i].module))
         load_output(&outputs[i]);
   }

   o = read_bufstream(b, &temp, sizeof(int));
   if(o != sizeof(int))
      out_set_active(0);
   else
      out_set_active(temp);
}

void load_filters(BUFSTREAM *b)
{
   int i, temp;
   int o = read_bufstream(b, &num_filters, sizeof(int));
   if(o != sizeof(int))
   {
      load_default_filters();
      return;
   }

   free(filters);
   filters = (FILTER *) malloc(num_filters * sizeof(FILTER));

   for(i = 0; i < num_filters; i++)
   {
      o = read_bufstream(b, &temp, sizeof(int));
      if(o != sizeof(int))
      {
         load_default_filters();
         return;
      }
      o = read_bufstream(b, &filters[i].module_name, temp);
      filters[i].module_name[temp] = 0;
      if(o != temp)
      {
         load_default_filters();
         return;
      }

      filters[i].enabled = FALSE;
      if(load_module(filters[i].module_name, &filters[i].module))
      {
         load_filter(&filters[i]);

         o = read_bufstream(b, &temp, sizeof(int));
         if(o != sizeof(int))
            filters[i].enabled = TRUE;
         else
            filters[i].enabled = temp;
      }
   }
}


BOOL save_decoders(BUFSTREAM *b)
{
   int i, temp;
   int o = write_bufstream(b, &num_decoders, sizeof(int));
   if(o != sizeof(int))
      return FALSE;

   for(i = 0; i < num_decoders; i++)
   {
      temp = strlen(decoders[i].module_name);

      o = write_bufstream(b, &temp, sizeof(int));
      if(o != sizeof(int))
         return FALSE;

      o = write_bufstream(b, &decoders[i].module_name, temp);
      if(o != temp)
         return FALSE;

      o = write_bufstream(b, &decoders[i].enabled, sizeof(int));
      if(o != sizeof(int))
         return FALSE;

   }
   return TRUE;
}

BOOL save_outputs(BUFSTREAM *b)
{
   int i, temp;
   int o = write_bufstream(b, &num_outputs, sizeof(int));
   if(o != sizeof(int))
      return FALSE;

   for(i = 0; i < num_outputs; i++)
   {
      temp = strlen(outputs[i].module_name);

      o = write_bufstream(b, &temp, sizeof(int));
      if(o != sizeof(int))
         return FALSE;

      o = write_bufstream(b, &outputs[i].module_name, temp);
      if(o != temp)
         return FALSE;
   }

   o = write_bufstream(b, &active_output, sizeof(int));
   if(o == sizeof(int))
      return TRUE;
   else
      return FALSE;
}

BOOL save_filters(BUFSTREAM *b)
{
   int i, temp;
   int o = write_bufstream(b, &num_filters, sizeof(int));
   if(o != sizeof(int))
      return FALSE;

   for(i = 0; i < num_filters; i++)
   {
      temp = strlen(filters[i].module_name);

      o = write_bufstream(b, &temp, sizeof(int));
      if(o != sizeof(int))
         return FALSE;

      o = write_bufstream(b, &filters[i].module_name, temp);
      if(o != temp)
         return FALSE;

      o = write_bufstream(b, &filters[i].enabled, sizeof(int));
      if(o != sizeof(int))
         return FALSE;

   }
   return TRUE;
}


#if 0

void load_decoders(FILE *file)
{
   int i, temp;
   int o = (int) fread(&num_decoders, 1, sizeof(int), file);
   if(o != sizeof(int))
   {
      load_default_decoders();
      return;
   }
   decoders = (DECODER *) malloc(num_decoders * sizeof(DECODER));

   for(i = 0; i < num_decoders; i++)
   {
      o = (int) fread(&temp, 1, sizeof(int), file);
      if(o != sizeof(int))
      {
         load_default_decoders();
         return;
      }
      o = (int) fread(&decoders[i].module_name, 1, temp, file);
      decoders[i].module_name[temp] = 0;
      if(o != temp)
      {
         load_default_decoders();
         return;
      }

      decoders[i].enabled = FALSE;
      if(load_module(decoders[i].module_name, &decoders[i].module))
      {
         load_decoder(&decoders[i]);

         o = (int) fread(&temp, 1, sizeof(int), file);
         if(o != sizeof(int))
            decoders[i].enabled = TRUE;
         else
            decoders[i].enabled = temp;
      }
   }
}


void load_outputs(FILE *file)
{
   int i, temp;
   int o = (int) fread(&num_outputs, 1, sizeof(int), file);
   if(o != sizeof(int))
   {
      load_default_outputs();
      return;
   }
   outputs = (OUTPUT *) malloc(num_outputs * sizeof(OUTPUT));

   for(i = 0; i < num_outputs; i++)
   {
      o = (int) fread(&temp, 1, sizeof(int), file);
      if(o != sizeof(int))
      {
         load_default_outputs();
         return;
      }
      o = (int) fread(&outputs[i].module_name, 1, temp, file);
      outputs[i].module_name[temp] = 0;
      if(o != temp)
      {
         load_default_outputs();
         return;
      }

      if(load_module(outputs[i].module_name, &outputs[i].module))
         load_output(&outputs[i]);
   }

   o = (int) fread(&temp, 1, sizeof(int), file);
   if(o != sizeof(int))
      out_set_active(0);
   else
      out_set_active(temp);
}

void load_filters(FILE *file)
{
   int i, temp;
   int o = (int) fread(&num_filters, 1, sizeof(int), file);
   if(o != sizeof(int))
   {
      load_default_filters();
      return;
   }
   filters = (FILTER *) malloc(num_filters * sizeof(FILTER));

   for(i = 0; i < num_filters; i++)
   {
      o = (int) fread(&temp, 1, sizeof(int), file);
      if(o != sizeof(int))
      {
         load_default_filters();
         return;
      }
      o = (int) fread(&filters[i].module_name, 1, temp, file);
      filters[i].module_name[temp] = 0;
      if(o != temp)
      {
         load_default_filters();
         return;
      }

      filters[i].enabled = FALSE;
      if(load_module(filters[i].module_name, &filters[i].module))
      {
         load_filter(&filters[i]);

         o = (int) fread(&temp, 1, sizeof(int), file);
         if(o != sizeof(int))
            filters[i].enabled = TRUE;
         else
            filters[i].enabled = temp;
      }
   }
}

BOOL save_decoders(FILE *file)
{
   int i, temp;
   int o = fwrite(&num_decoders, 1, sizeof(int), file);
   if(o != sizeof(int))
      return FALSE;

   for(i = 0; i < num_decoders; i++)
   {
      temp = strlen(decoders[i].module_name);

      o = (int) fwrite(&temp, 1, sizeof(int), file);
      if(o != sizeof(int))
         return FALSE;

      o = (int) fwrite(&decoders[i].module_name, 1, temp, file);
      if(o != temp)
         return FALSE;

      o = (int) fwrite(&decoders[i].enabled, 1, sizeof(int), file);
      if(o != sizeof(int))
         return FALSE;

   }
   return TRUE;
}

BOOL save_outputs(FILE *file)
{
   int i, temp;
   int o = fwrite(&num_outputs, 1, sizeof(int), file);
   if(o != sizeof(int))
      return FALSE;

   for(i = 0; i < num_outputs; i++)
   {
      temp = strlen(outputs[i].module_name);

      o = (int) fwrite(&temp, 1, sizeof(int), file);
      if(o != sizeof(int))
         return FALSE;

      o = (int) fwrite(&outputs[i].module_name, 1, temp, file);
      if(o != temp)
         return FALSE;
   }

   o = (int) fwrite(&active_output, 1, sizeof(int), file);
   if(o == sizeof(int))
      return TRUE;
   else
      return FALSE;
}

BOOL save_filters(FILE *file)
{
   int i, temp;
   int o = fwrite(&num_filters, 1, sizeof(int), file);
   if(o != sizeof(int))
      return FALSE;

   for(i = 0; i < num_filters; i++)
   {
      temp = strlen(filters[i].module_name);

      o = (int) fwrite(&temp, 1, sizeof(int), file);
      if(o != sizeof(int))
         return FALSE;

      o = (int) fwrite(&filters[i].module_name, 1, temp, file);
      if(o != temp)
         return FALSE;

      o = (int) fwrite(&filters[i].enabled, 1, sizeof(int), file);
      if(o != sizeof(int))
         return FALSE;

   }
   return TRUE;
}

#endif


// returns the type of plug-ins contained in this module
ULONG check_plugin(HMODULE module, char *module_name, PLUGIN_QUERYPARAM *query_param)
{
   void (* _System plugin_query)(PLUGIN_QUERYPARAM *param);

   if(load_function(module, (PFN*)&plugin_query, "plugin_query", module_name))
   {
      (* plugin_query)(query_param);
      return query_param->type;
   }
   else
      return 0;
}

BOOL add_decoder_plugin(DECODER *decoder)
{
   int i;
   for(i = 0; i < num_decoders; i++)
   {
      if(decoders[i].module == decoder->module)
         return FALSE; // already loaded
   }

   load_decoder(decoder);
   num_decoders++;
   decoders = (DECODER *) realloc(decoders, num_decoders * sizeof(DECODER));
   decoders[num_decoders-1] = *decoder;
   return TRUE; // if functions didn't load, it's the user's problem
}


BOOL add_output_plugin(OUTPUT *output)
{
   int i;
   for(i = 0; i < num_outputs; i++)
   {
      if(outputs[i].module == output->module)
         return FALSE; // already loaded
   }

   load_output(output);

   num_outputs++;
   outputs = (OUTPUT *) realloc(outputs, num_outputs * sizeof(OUTPUT));
   outputs[num_outputs-1] = *output;
   if(active_output == -1)
      out_set_active(num_outputs-1);
   return TRUE; // if functions didn't load, it's the user's problem
}


BOOL add_filter_plugin(FILTER *filter)
{
   int i;
   for(i = 0; i < num_filters; i++)
   {
      if(filters[i].module == filter->module)
         return FALSE; // already loaded
   }

   load_filter(filter);
   num_filters++;
   filters = (FILTER *) realloc(filters, num_filters * sizeof(FILTER));
   filters[num_filters-1] = *filter;
   return TRUE; // if functions didn't load, it's the user's problem
}



//returns the types found
ULONG add_plugin(char *module_name)
{
   HMODULE module;
   if(load_module(module_name, &module))
   {
      PLUGIN_QUERYPARAM param;
      int types = check_plugin(module, module_name, &param);

      if(types & PLUGIN_VISUAL)
         return FALSE;
      if(types & PLUGIN_FILTER)
      {
         FILTER filter;
         strcpy(filter.module_name, module_name);
         filter.module = module;

         if(!add_filter_plugin(&filter))
            types = types & ~PLUGIN_FILTER;
      }

      if(types & PLUGIN_DECODER)
      {
         DECODER decoder;
         strcpy(decoder.module_name, module_name);
         decoder.module = module;

         if(!add_decoder_plugin(&decoder))
            types = types & ~PLUGIN_DECODER;
      }

      if(types & PLUGIN_OUTPUT)
      {
         OUTPUT output;
         strcpy(output.module_name, module_name);
         output.module = module;

         if(!add_output_plugin(&output))
            types = types & ~PLUGIN_OUTPUT;
      }
      return types;
   }
   return 0;
}

BOOL remove_decoder_plugin(DECODER *decoder)
{
   int i;
   BOOL found = FALSE;
   for(i = 0; i < num_decoders && !found; i++)
   {
      if(&decoders[i] == decoder)
         found = TRUE;
   }
   if(!found)
      return FALSE;

   i--;

   if(active_decoder == i)
      dec_set_active(-1);

   unload_module(decoders[i].module);
   if(i<num_decoders-1)
      memmove(&decoders[i],&decoders[i+1],(num_decoders-i-1)*sizeof(DECODER));
   num_decoders--;
   decoders = (DECODER *) realloc(decoders,num_decoders*sizeof(DECODER));
   return TRUE;
}

BOOL remove_output_plugin(OUTPUT *output)
{
   int i;
   BOOL found = FALSE;
   for(i = 0; i < num_outputs && !found; i++)
   {
      if(&outputs[i] == output)
         found = TRUE;
   }
   if(!found)
      return FALSE;

   i--;

   if(active_output == i)
      out_set_active(-1);

   unload_module(output[i].module);
   if(i<num_outputs-1)
      memmove(&outputs[i],&outputs[i+1],(num_outputs-i-1)*sizeof(OUTPUT));
   num_outputs--;
   outputs = (OUTPUT *) realloc(outputs,num_outputs*sizeof(OUTPUT));
   return TRUE;
}

BOOL remove_filter_plugin(FILTER *filter)
{
   int i;
   BOOL found = FALSE;
   for(i = 0; i < num_filters && !found; i++)
   {
      if(&filters[i] == filter)
         found = TRUE;
   }
   if(!found)
      return FALSE;

   i--;

   filter->filter_uninit(filter->f);
   filter->f = NULL;

   unload_module(filters[i].module);
   if(i<num_filters-1)
      memmove(&filters[i],&filters[i+1],(num_filters-i-1)*sizeof(FILTER));
   num_filters--;
   filters = (FILTER *) realloc(filters,num_filters*sizeof(FILTER));
   return TRUE;
}



// returns -1 if error, -2 can't find nothing, otherwize may be the decoder's thread
int dec_set_name_active(char *name)
{
   int i;
   if(name == NULL || name[0] == 0)
      return dec_set_active(-1);

   for(i = 0; i < num_decoders; i++)
   {
      char temp[256];
      fnameext(temp, decoders[i].module_name);
      if(stricmp(temp,name) == 0)
         return dec_set_active(i);
   }

   return -2;
}

// returns -2 not enabled, returns -1 if error, otherwize may be the decoder's thread
int dec_set_active(int number)
{
   if(number >= num_decoders || number < -1)
      return -1;

   if(active_decoder == number)
      return 0;

   if(active_decoder != -1)
   {
      decoders[active_decoder].decoder_uninit(decoders[active_decoder].w);
      decoders[active_decoder].w = NULL;
   }

   if(number != -1 && !decoders[number].enabled)
      return -2;

   active_decoder = number;

   if(active_decoder != -1)
      return decoders[active_decoder].decoder_init(&decoders[active_decoder].w);
   else
      return 0;
}

// 0 = ok, 1 = command unsupported, 3 = no decoder active, 4 = no active output, others unimportant
ULONG _System dec_command(ULONG msg, DECODER_PARAMS *params)
{
   if(active_decoder != -1)
   {
      // is it good to setup output_play_samples and filter chain only here?
      if(msg == DECODER_SETUP)
      {
         if(active_output == -1)
         {
            error_display("There is no active output.");
            return 4;
         }
         else
         {
            FILTER **enabled_filters = (FILTER**) alloca(num_filters * sizeof(FILTER*));
            int num_enabled_filters = 0;
            int i = 0;

            for(i = 0; i < num_filters; i++)
               if(filters[i].enabled)
                  enabled_filters[num_enabled_filters++] = &filters[i];

            if(num_enabled_filters > 0)
            {
               FILTER_PARAMS filter_params;
               i = num_enabled_filters-1;

               // you need to start from the last filter to init them all properly
               if(enabled_filters[i]->f != NULL)
               {
                  enabled_filters[i]->filter_uninit(enabled_filters[i]->f);
                  enabled_filters[i]->f = NULL;
               }

               filter_params.error_display = params->error_display;
               filter_params.info_display = params->error_display;
               filter_params.output_play_samples = outputs[active_output].output_play_samples;
               filter_params.a = outputs[active_output].a;
               filter_params.audio_buffersize = params->audio_buffersize;
               enabled_filters[i]->filter_init(&enabled_filters[i]->f,&filter_params);

               for(i = num_enabled_filters-2; i >= 0; i--)
               {
                  if(enabled_filters[i]->f != NULL)
                  {
                     enabled_filters[i]->filter_uninit(enabled_filters[i]->f);
                     enabled_filters[i]->f = NULL;
                  }

                  filter_params.error_display = params->error_display;
                  filter_params.info_display = params->info_display;
                  filter_params.output_play_samples = enabled_filters[i+1]->filter_play_samples;
                  filter_params.a = enabled_filters[i+1]->f;
                  filter_params.audio_buffersize = params->audio_buffersize;
                  enabled_filters[i]->filter_init(&enabled_filters[i]->f,&filter_params);
               }

               params->output_play_samples = enabled_filters[0]->filter_play_samples;
               params->a = enabled_filters[0]->f;
            }
            else
            {
               params->output_play_samples = outputs[active_output].output_play_samples;
               params->a = outputs[active_output].a;
            }

         }
      }
//   printf("dec_command %d %d %d\n",msg, active_output,outputs[active_output].a);
      return decoders[active_decoder].decoder_command(decoders[active_decoder].w,msg,params);
   }
   else
      return 3; // no decoder active
}

// returns the decoder NAME that can play this file and returns 0
// if not returns error 100 = error reading, 200 = nothing can play that,
// 1001 = http error, other values = errno
ULONG _System dec_fileinfo(char *filename, DECODER_INFO *info, char *name)
{
   ULONG last_rc = 200;
   int i;
   for(i = 0; i < num_decoders; i++)
   {
      if(decoders[i].enabled && decoders[i].decoder_fileinfo != NULL)
      {
         last_rc = decoders[i].decoder_fileinfo(filename,info);
         if(last_rc == 0 && name != NULL)
            fnameext(name,decoders[i].module_name);
         if(last_rc == 0)
            return 0;
      }
   }

   return last_rc; // returns only the last RC ... hum
}

ULONG _System dec_trackinfo(char *drive, int track, DECODER_INFO *info, char *name)
{
   ULONG last_rc = 200;
   int i;
   for(i = 0; i < num_decoders; i++)
   {
      if(decoders[i].enabled && decoders[i].decoder_trackinfo != NULL)
      {
         last_rc = decoders[i].decoder_trackinfo(drive,track,info);
         if(last_rc == 0)
         {
            fnameext(name,decoders[i].module_name);
            return 0;
         }
      }
   }

   return last_rc; // returns only the last RC ... hum
}

ULONG _System dec_cdinfo(char *drive, DECODER_CDINFO *info)
{
   ULONG last_rc = 200;
   int i;
   for(i = 0; i < num_decoders; i++)
   {
      if(decoders[i].enabled && decoders[i].decoder_cdinfo != NULL)
      {
         last_rc = decoders[i].decoder_cdinfo(drive,info);
         if(last_rc == 0)
            return 0;
      }
   }

   return last_rc; // returns only the last RC ... hum
}



// DECODER_STOPPED  0
// DECODER_PLAYING  1
// DECODER_STARTING 2
// no decoder active 3

ULONG _System dec_status()
{
   if(active_decoder != -1)
      return decoders[active_decoder].decoder_status(decoders[active_decoder].w);
   else
      return 3; // no decoder active
}

// length in ms, should still be valid if decoder stops
ULONG _System dec_length()
{
   if(active_decoder != -1)
      return decoders[active_decoder].decoder_length(decoders[active_decoder].w);
   else
      return 0; // bah??
}


int out_set_active(int number)
{
//   printf("%d %d %d\n",number,active_output,outputs[active_output].a);
   if(number >= num_outputs || number < -1)
      return -1;

   if(active_output == number)
      return 0;

   if(active_output != -1)
   {
      outputs[active_output].output_command(outputs[active_output].a,OUTPUT_CLOSE,NULL);
      outputs[active_output].output_uninit(outputs[active_output].a);
      outputs[active_output].a = NULL;
   }

   active_output = number;

   if(active_output != -1)
   {
      outputs[active_output].output_init(&outputs[active_output].a);
//   printf("bam %d %d %d\n",number,active_output,outputs[active_output].a);
      return 0;
   }
   else
      return 0;
}

// returns -2 can't find nothing
int out_set_name_active(char *name)
{
   int i;

   if(name == NULL || name[0] == 0)
      return out_set_active(-1);

   for(i = 0; i < num_outputs; i++)
   {
      char temp[256];
      fnameext(temp, outputs[i].module_name);
      if(stricmp(temp,name) == 0)
      {
         out_set_active(i);
         return 0;
      }
   }

   return -2;
}


// 0 = ok, else = return code from MMOS/2
ULONG out_command(ULONG msg, OUTPUT_PARAMS *ai)
{
   if(active_output != -1)
   {
//   printf("out_command %d %d\n",active_output,outputs[active_output].a);
      outputs[active_output].output_command(outputs[active_output].a,msg,ai);
//   printf("cooked\n");
   return 0;

   }
   else
   {
      error_display("There is no active output plug-in.");
      return -1; // ??
   }
}


// returns 0 = success otherwize MMOS/2 error
ULONG _System out_playing_samples(FORMAT_INFO *info, char *buf, int len)
{
   if(active_output != -1)
      return outputs[active_output].output_playing_samples(outputs[active_output].a,info,buf,len);
   else
      return -1; // ??
}

// returns time in ms
ULONG _System out_playing_pos()
{
   if(active_output != -1)
      return outputs[active_output].output_playing_pos(outputs[active_output].a);
   else
      return 0; // ??
}

// if the output is playing
BOOL _System out_playing_data()
{
   if(active_output != -1)
      return outputs[active_output].output_playing_data(outputs[active_output].a);
   else
      return FALSE;
}


/* backward compatibility */
BOOL decoder_playing()
{
   ULONG status = dec_status();
   return (status == DECODER_PLAYING || status == DECODER_STARTING || out_playing_data());
}

#define TOTALSECS  1  /* returns less than 0 if unknown */
#define PLAYEDSECS 2

ULONG decoder_time(ULONG which)
{
   if(which == TOTALSECS)
      return dec_length()/1000;
   else
      return out_playing_pos()/1000;
}


/* plug-in menu in the main pop-up menu */

typedef struct
{
   char *filename;
   int type;
   int i;
   int configurable;
   int enabled;

} PLUGIN_ENTRY;

typedef struct
{
   PLUGIN_ENTRY *entries;
   int num_entries;
} PLUGINS;

PLUGINS mpi = {0}; // plug-in menu entries

extern PLUGIN plugins[64]; // old visual crap...


void load_plugin_menu(HWND hMenu)
{
   char temp[512];
   char buffer[2048];
   MENUITEM mi;
   int i = 0;
   int count;
   int num_plugins = 0;
   int next_entry = 0;

   // delete all
   count = LONGFROMMR(WinSendMsg(hMenu, MM_QUERYITEMCOUNT, 0, 0));
   for(i = 0; i < count; i++)
   {
      short item = LONGFROMMR(WinSendMsg(hMenu, MM_ITEMIDFROMPOSITION, MPFROMSHORT(0), 0));
      WinSendMsg(hMenu,MM_DELETEITEM, MPFROM2SHORT(item, FALSE), 0);
   }

   for(i = 0; i < mpi.num_entries; i++)
   {
      free(mpi.entries[i].filename);
   }
   free(mpi.entries);
   mpi.entries = NULL;
   mpi.num_entries = 0;

   // visual plug-ins
   for (i = 0; i < 32; i++)
   {
      if (strcmp(plugins[i].filename, "") != 0)
      {
         num_plugins++;
      }
   }

   num_plugins += num_decoders + num_outputs + num_filters;

   if(num_plugins == 0)
   {
      mi.iPosition = MIT_END;
      mi.afStyle = MIS_TEXT;
      mi.afAttribute = MIA_DISABLED;
      mi.id = IDM_M_BOOKMARKS + 1;
      mi.hwndSubMenu = (HWND)NULLHANDLE;
      mi.hItem = 0;
      WinSendMsg(hMenu, MM_INSERTITEM, MPFROMP(&mi), MPFROMP("No plug-ins"));

      goto end;
   }

   mpi.num_entries = num_plugins;
   mpi.entries = (PLUGIN_ENTRY *) malloc(mpi.num_entries * sizeof(PLUGIN_ENTRY));

   // visual plug-ins
   for (i = 0; i < 32; i++)
   {
      if (strcmp(plugins[i].filename, "") != 0)
      {
         sprintf(buffer, "%s (%s%s)", plugins[i].desc,sfname(plugins[i].filename), sext(plugins[i].filename));
         mpi.entries[next_entry].filename = strdup(buffer);
         mpi.entries[next_entry].type = PLUGIN_VISUAL;
         mpi.entries[next_entry].i = i;
         mpi.entries[next_entry].configurable = plugins[i].configurable;
         mpi.entries[next_entry].enabled = plugins[i].loaded;
         next_entry++;
      }
   }

   // decoder plug-ins
   for (i = 0; i < num_decoders; i++)
   {
      fnameext(temp, decoders[i].module_name);
      sprintf(buffer, "%s (%s)", decoders[i].query_param.desc, temp);
      mpi.entries[next_entry].filename = strdup(buffer);
      mpi.entries[next_entry].type = PLUGIN_DECODER;
      mpi.entries[next_entry].i = i;
      mpi.entries[next_entry].configurable = decoders[i].query_param.configurable;
      mpi.entries[next_entry].enabled = decoders[i].enabled;
      next_entry++;
   }

   // output plug-ins
   for (i = 0; i < num_outputs; i++)
   {
      fnameext(temp, outputs[i].module_name);
      sprintf(buffer, "%s (%s)", outputs[i].query_param.desc, temp);
      mpi.entries[next_entry].filename = strdup(buffer);
      mpi.entries[next_entry].type = PLUGIN_OUTPUT;
      mpi.entries[next_entry].i = i;
      mpi.entries[next_entry].configurable = outputs[i].query_param.configurable;
      mpi.entries[next_entry].enabled = (i == active_output ? TRUE : FALSE);
      next_entry++;
   }

   // filter plug-ins
   for (i = 0; i < num_filters; i++)
   {
      fnameext(temp, filters[i].module_name);
      sprintf(buffer, "%s (%s)", filters[i].query_param.desc, temp);
      mpi.entries[next_entry].filename = strdup(buffer);
      mpi.entries[next_entry].type = PLUGIN_FILTER;
      mpi.entries[next_entry].i = i;
      mpi.entries[next_entry].configurable = filters[i].query_param.configurable;
      mpi.entries[next_entry].enabled = filters[i].enabled;
      next_entry++;
   }

   for(i = 0; i < mpi.num_entries; i++)
   {
      mi.iPosition = MIT_END;
      mi.afStyle = MIS_TEXT;
      mi.afAttribute = 0;
      if(!mpi.entries[i].configurable)
         mi.afAttribute |= MIA_DISABLED;
      if(mpi.entries[i].enabled)
         mi.afAttribute |= MIA_CHECKED;
      mi.id = IDM_M_PLUG + i + 1;
      mi.hwndSubMenu = (HWND)NULLHANDLE;
      mi.hItem = 0;
      WinSendMsg(hMenu, MM_INSERTITEM, MPFROMP(&mi), MPFROMP(mpi.entries[i].filename));
   }

end:

   return;
}

extern HWND hwndFrame; // main window frame

BOOL process_possible_plugin(USHORT cmd)
{
   int i = cmd-IDM_M_PLUG-1;
   if(i >= 0 && i < mpi.num_entries)
   {
      switch(mpi.entries[i].type)
      {
         case PLUGIN_VISUAL:
            cfg_plugin(hwndFrame, mpi.entries[i].i);
            break;
         case PLUGIN_DECODER:
            decoders[mpi.entries[i].i].plugin_configure(hwndFrame, decoders[mpi.entries[i].i].module);
            break;
         case PLUGIN_OUTPUT:
            outputs[mpi.entries[i].i].plugin_configure(hwndFrame, outputs[mpi.entries[i].i].module);
            break;
         case PLUGIN_FILTER:
            filters[mpi.entries[i].i].plugin_configure(hwndFrame, filters[mpi.entries[i].i].module);
            break;
      }
      return TRUE;
   }
   else
      return FALSE;
}

