/*
 * Copyright 1997-2003 Samuel Audet <guardia@step.polymtl.ca>
 *                     Taneli Lepp <rosmo@sektori.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


typedef struct
{
   HMODULE module;
   char module_name[128];
   PLUGIN_QUERYPARAM query_param;
   BOOL enabled;

   void *w;
   int  (* _System decoder_init)(void **w);
   BOOL (* _System decoder_uninit)(void *w);
   ULONG (* _System decoder_command)(void *w, ULONG msg, DECODER_PARAMS *params);
   ULONG (* _System decoder_status)(void *w);
   ULONG (* _System decoder_length)(void *w);
   ULONG (* _System decoder_fileinfo)(char *filename, DECODER_INFO *info);
   ULONG (* _System decoder_trackinfo)(char *drive, int track, DECODER_INFO *info);
   ULONG (* _System decoder_cdinfo)(char *drive, DECODER_CDINFO *info);

   void (* _System plugin_query)(PLUGIN_QUERYPARAM *param);
   void (* _System plugin_configure)(HWND hwnd, HMODULE module);

} DECODER;

typedef struct
{
   HMODULE module;
   char module_name[128];
   PLUGIN_QUERYPARAM query_param;

   void *a;
   ULONG (* _System output_init)(void **a);
   ULONG (* _System output_uninit)(void *a);
   ULONG (* _System output_command)(void *a, ULONG msg, OUTPUT_PARAMS *info);
   ULONG (* _System output_playing_samples)(void *a, FORMAT_INFO *info, char *buf, int len);
   int (* _System output_play_samples)(void *a, FORMAT_INFO *format, char *buf,int len, int posmarker);
   int (* _System output_playing_pos)(void *a);
   BOOL (* _System output_playing_data)(void *a);

   void (* _System plugin_query)(PLUGIN_QUERYPARAM *param);
   void (* _System plugin_configure)(HWND hwnd, HMODULE module);

} OUTPUT;

typedef struct
{
   HMODULE module;
   char module_name[128];
   PLUGIN_QUERYPARAM query_param;
   BOOL enabled;

   void *f;
   ULONG (* _System filter_init)(void **f, FILTER_PARAMS *params);
   BOOL (* _System filter_uninit)(void *f);
   int (* _System filter_play_samples)(void *f, FORMAT_INFO *format, char *buf,int len, int posmarker);

   void (* _System plugin_query)(PLUGIN_QUERYPARAM *param);
   void (* _System plugin_configure)(HWND hwnd, HMODULE module);

} FILTER;


// these externs are not supposed to be used to make stupid things,
// but only to READ for configuration purposes, or set enabled flag
extern DECODER *decoders;
extern int num_decoders;
extern int active_decoder;
extern OUTPUT *outputs;
extern int num_outputs;
extern int active_output;
extern FILTER *filters;
extern int num_filters;


BOOL save_decoders(BUFSTREAM *b);
BOOL save_outputs(BUFSTREAM *b);
BOOL save_filters(BUFSTREAM *b);
void load_decoders(BUFSTREAM *b);
void load_outputs(BUFSTREAM *b);
void load_filters(BUFSTREAM *b);

void load_default_decoders(void);
void load_default_outputs(void);
void load_default_filters(void);

ULONG add_plugin(char *module_name);
BOOL remove_decoder_plugin(DECODER *decoder);
BOOL remove_output_plugin(OUTPUT *output);
BOOL remove_filter_plugin(FILTER *filter);

int dec_set_name_active(char *name);
int dec_set_active(int number);
ULONG _System dec_command(ULONG msg, DECODER_PARAMS *params);
ULONG _System dec_fileinfo(char *filename, DECODER_INFO *info, char *name);
ULONG _System dec_trackinfo(char *drive, int track, DECODER_INFO *info, char *name);
ULONG _System dec_cdinfo(char *drive, DECODER_CDINFO *info);
ULONG _System dec_status(void);
ULONG _System dec_length(void);

int out_set_name_active(char *name);
int out_set_active(int number);
ULONG _System out_command(ULONG msg, OUTPUT_PARAMS *info);
ULONG _System out_playing_samples(FORMAT_INFO *info, char *buf, int len);
ULONG _System out_playing_pos(void);
BOOL _System out_playing_data(void);

/* backward compatibility */
BOOL _System decoder_playing(void);
ULONG _System decoder_time(ULONG which);

/* plug-in menu in the main pop-up menu */
void load_plugin_menu(HWND hMenu);
BOOL process_possible_plugin(USHORT cmd);

