/*
 * Copyright 1997-2003 Samuel Audet <guardia@step.polymtl.ca>
 *                     Taneli Lepp <rosmo@sektori.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "tag.h"
#include "iniman.h"
#include "gbmv.h"
#include <time.h>
#include <sys/time.h>

#define VERSION "PM123 1.21"

#define NOTEBOOK_ACCEL     8000

#define MSG_SCROLL            1
#define MSG_BLINK             2

#define MSG_PLAY              1
#define MSG_STOP              2
#define MSG_PRIO              3
#define MSG_PAUSE             4
#define MSG_FWD               5
#define MSG_REW               6
#define MSG_JUMP              7
#define MSG_SAVE              8

#define PLAYLIST_MENU        900
#define IDM_PL_LOAD          901
#define IDM_PL_REMOVE        902
#define IDM_PL_SORT_SIZE     903
#define IDM_PL_SORT_PLTIME   904
#define IDM_PL_SORT_NAME     905
#define IDM_PL_USE           906
#define IDM_PL_CLEAR         907
#define IDM_PL_SAVE          908
#define IDM_PL_LOADL         909
#define IDM_PL_SORT_SONG     910
#define IDM_PL_SORT_RANDOM   970
#define IDM_PL_MENU          950
#define IDM_PL_LAST          951
#define IDM_PL_URL           965
#define IDM_PL_MENUCONT      971
#define IDM_PL_TRACK         972


#define PLAYLISTS_MENU       910
#define IDM_PL_S_PLAY        911
#define IDM_PL_S_DEL         912
#define IDM_PL_S_TAG         913
#define IDM_PL_S_DTAG        914


#define MAIN_MENU            500
#define IDM_M_PRIDLE         501
#define IDM_M_PRREG          502
#define IDM_M_PRFG           503
#define IDM_M_PRCRIT         504
#define IDM_M_SHUFFLE        505
#define IDM_M_ABOUT          506
#define IDM_M_CFG            507
#define IDM_M_LOAD           508
#define IDM_M_PLAYLIST       509
#define IDM_M_MINIMIZE       510
#define IDM_M_SMALL          511
#define IDM_M_TINY           512
#define IDM_M_NORMAL         513
#define IDM_M_URL            514
#define IDM_M_FONT1          515
#define IDM_M_FONT2          516
#define IDM_M_PRLOW          517
#define IDM_M_PRHIGH         518
#define IDM_M_FONT3          519
#define IDM_M_SKIN           520
#define IDM_M_FLOAT          521
#define IDM_M_EQUALIZE       522
#define IDM_M_TAG            523
#define IDM_M_LAST           524 // 10 IDs over IDM_M_BOOKMARKS need to be free
#define IDM_M_MANAGER        540
#define IDM_M_TRACK          541
#define IDM_M_SAVE           542

#define IDM_M_ADDBOOK        545
#define IDM_M_EDITBOOK       546
#define IDM_M_BOOKMARKS      600 // a lot of IDs after IDM_M_BOOKMARKS need to be free
#define IDM_M_PLUG           7000 // a lot of IDs after IDM_M_PLUG need to be free



#define DIRECTORY_MENU       550
#define IDM_DIR_RECURSE      551
#define IDM_DIR_SINGLE       552
#define IDM_DIR_CANCEL       553

#define AMP_NOFILE             0
#define AMP_SINGLE             1
#define AMP_PLAYLIST           2

#define DIGIT_SMALL         1100
#define DIGIT_BIG           1200
#define DIGIT_TINY          1640

#define LISTDLG               42
#define LIST_STATIC          666
#define ABOUTDLG              46
#define PMDLG                 48
#define BOOKMARK_DLG          49

#define DIGIT_DOTDOT          10

#define BMP_SINGLEPLAY      1631
#define BMP_LISTPLAY        1632
#define BMP_NOFILE          1633

/* Bitmap handles for action buttons */
#define BMP_PLAY    1300
#define BMP_STOP    1323
#define BMP_PAUSE   1302
#define BMP_REW     1303
#define BMP_FWD     1304
#define BMP_POWER   1305
#define BMP_PREV    1306
#define BMP_NEXT    1307
#define BMP_LIST    1308
#define BMP_LOAD    1309
#define BMP_N_PLAY  1310
#define BMP_N_STOP  1322
#define BMP_N_PAUSE 1312
#define BMP_N_REW   1313
#define BMP_N_FWD   1314
#define BMP_N_POWER 1315
#define BMP_N_PREV  1316
#define BMP_N_NEXT  1317
#define BMP_N_LIST  1318
#define BMP_N_LOAD  1319
#define BMP_N_PL    1320
#define BMP_PL      1321
#define BMP_FLOAD   1325
#define BMP_N_FLOAD 1324

#define BMP_KILL    1600
#define BMP_STEREO  1601
#define BMP_MONO    1602
#define BMP_VOL     1603
#define BMP_NOVOL   1604
#define BMP_SCOPE   1605
#define BMP_NOSCOPE 1606
#define BMP_SDGLOGO 1607
#define BMP_VOLBAR  1630

#define BMP_BARS    1610

#define BMP_LED     1820
#define BMP_NOLED   1821

#define BMP_NOTL    1901
#define BMP_TL      1902
#define BMP_NOPLIST 1903
#define BMP_PLIST   1904

/* Container */
typedef struct _MYMPREC {
  RECORDCORE rc;
  HPOINTER   icon;
  PCHAR      songname;
  PCHAR      length;
  PCHAR      time;
  PCHAR      info;
  PCHAR      full;
  int        played;
  int        bitrate, channels;
  int        secs;
  int        mode, freq;
  ULONG      size;

  char       decoder_module_name[16];
  FORMAT_INFO format;
  char *info_string;
  char track;
  char cd_drive[4];
  char *comment;
} MYMPREC, *PMYMPREC;

/* for PM123.INI */
struct amp_cfg {
  char id[48];
  char eof;

  char filedir[CCHMAXPATH];
  char listdir[CCHMAXPATH];
  char savedir[CCHMAXPATH];
  char cddrive[4];
  char lasteq[CCHMAXPATH];
  int eq_enabled;

  int defaultvol;
  int playonload;               /* 0=don't, 1=do        */
  int autouse;                  /* 0=don't, 1=do        */
  char pl_font[128];            /* playlist font        */
  RGB2 pl_fg_color;             /* pleilist kolor f0re  */
  RGB2 pl_bg_color;             /* pleilist kolor bck  */
  int divider;                  /* xsplitbar crap       */
  int mode;                     /* 0=reg,1=small,2=tiny */
  int font;                     /* gui font             */
  int priority;                 /* 0=low 1=high         */
  int trash;                    /* trash bufs on seek   */
  char defaultskin[256];        /* default skin         */
  int shf, rpt;                 /* button statez        */
  int floatontop;               /* float on top         */
  int playonuse;                /* play on use playlist */
  int scroll;                   /* 0=inf, 1=once, 2=no  */
  char last[10][256];           /* last used files      */
  char list[10][256];           /* last used playlists  */
  time_t first_run;             /* first time ran       */
  int viewmode;                 /* def display m0de     */
  char proxy[1024];             /* proxy url            */
  char auth[256];               /* http authorization   */
  int bufwait;                  /* wait before fucking  */
  int bufsize;                  /* kb chunkz rewl       */

  SWP Main;
  SWP Playlist;
  SWP Manager;                  /* pl.manager window    */
  SWP Equalizer;                /* eq window            */
  SWP Bookmark;                 /* bookmark window      */
  SWP Properties;

  int show_playlist;            /* 1 = show, 0 = don't  */

  int dock_windows;             /* dock windows?        */
  int dock_margin;              /* the marging for docking window */

} cfg, oldcfg;

extern HMODULE thisModule;
extern char *thisModulePath;

extern float gains[20];
extern BOOL  mutes[20];
extern float preamp;

/* pm123.c */
void amp_load_visual_plugin(HWND hwnd);
BOOL amp_load_eq_file(char *filename, float *gains, BOOL *mutes, float *preamp);
void volnorm(void);

/* 123_util.c Util functions */
char *sdrive(char *buffer);
char *sdir(char *buffer);
char *sfname(char *buffer);
char *sext(char *buffer);
void fnameext(char *buf, char *fullpath);
void drivedir(char *buf, char *fullpath);
int count_files(int argc, char *argv[]);

int amp_gettag(char *fiilunimi, DECODER_INFO *info, tune *joku);
int amp_wipetag(char *fiilunimi);

void amp_movewindow(HWND hwnd);

int bmp_y(int res);
int bmp_x(int res);
int char_width(void);
int char_height(void);
void do_display_msg(char *filename);
void next_display_mode(void);
HPAL amp_dopalette(HAB hab, HWND hwnd);
void amp_thick3d(HPS hps, int x, int y, int width, int height, long clr1, long clr2);
void amp_drawdigit(HPS hps, int x, int y, int digit, int size);
HPS amp_beginpaint(HWND hwnd);
void amp_endpaint(HPS hps);
void amp_createmainwin(HWND hwnd);
void amp_box3d(HPS hps, int x, int y, int width, int height, long clr1, long clr2);
void amp_drawbackground(HPS hps, HWND hwnd, ULONG color);
int amp_drawstring(HPS hps, int x, int y, char *string, int maxlen);
void amp_listwindow(HWND hwnd);
HWND amp_listcontainer(HWND hwnd);
int amp_drop_listwin(HWND hwnd, MPARAM mp1);
HWND amp_drop_mainwin(HWND hwnd, MPARAM mp1, char *single, int size);
MRESULT amp_accept_drop(MPARAM mp1);
HWND amp_list_killbtn(HWND hwnd);
HWND amp_list_static(HWND hwnd);
char *amp_fsize(char *buffer);
void amp_resizelist(HWND hwnd, HWND container, HWND kill, HWND staattinen);
void amp_spectrumbar(HPS hps, int x, int y, int power);
void amp_volumebar(HWND hwnd, HPS hps, int x, int y, int height, int volume);
void amp_priochange(HWND hwndMenu, int newprio);
void amp_drawrate(HPS hps, int x, int y, int rate);
void amp_drawchannels(HPS hps, int x, int y, int channels);
int amp_addplaylist(HWND container, HWND hwnd, char *filename, int clear);
void amp_drawled(HPS hps, int x, int y, int state);
void amp_drawtimer(HPS hps, int x, int y, long time);
void amp_box(HPS hps, int x, int y, int width, int height, long clr1);
void set_bubbletext(USHORT id, char *text);
void lamepad(char *target, int width, char *string);
void amp_drawplind(HPS hps, int index, int total);
void amp_drawbmp(HPS hps, int x, int y, int res);
void amp_drawtinytimer(HPS hps, int x, int y, long time);
void amp_drawindicator(HPS hps, int position);
void amp_drawbk(HPS hps, int x1, int y1, int x2, int y2, int bground);

int init_skins(char *pfilename, HWND hwnd, HPS hps);

int find_index(PMYMPREC position);
PMYMPREC pl_add_song(char *filename, PMYMPREC position, char *title);
void reflow_and_resize(void);
char *construct_tag_info_string(tune *joopajoo);

int amp_puttag(char *fiilunimi, tune *joku);

void pipe_write(HPIPE hpipe, char *);
int recursive_dir_add(char *directory);
void playlist_post_file(char *pfilename);
void load_playlist(HWND container, HWND hwnd, char *pfilename, int clear);
int is_playlist(char *filename);
void free_all_records(HWND container);

SHORT _System pfnCompareSize(PMYMPREC p1, PMYMPREC p2, PVOID pStorage);
SHORT _System pfnCompareAlphabet(PMYMPREC p1, PMYMPREC p2, PVOID pStorage);
SHORT _System pfnCompareSongname(PMYMPREC p1, PMYMPREC p2, PVOID pStorage);
SHORT _System pfnCompareTime(PMYMPREC p1, PMYMPREC p2, PVOID pStorage);
SHORT _System pfnCompareRandom(PMYMPREC p1, PMYMPREC p2, PVOID pStorage);

int load_singlefile(char *file, HWND container, HWND hwnd);
BOOL load_singletrack(char *drive, char track);

typedef struct {
  char    filename[256];
  char    *author, *desc;
  int     type;
  int     x, y, cx, cy;
  HMODULE module;
  HWND    hwnd;
  HWND    hwndConfig;
  int     loaded;
  int     configurable;
  char    param[256];
  PFN     deinit;
} PLUGIN, *PPLUGIN;

ULONG handle_dfi_error(ULONG rc, char *file);

int init_plugin(HWND hwnd, int n);
void cfg_plugin(HWND hwnd, int n);
void deinit_plugin(HWND hwnd, int n);
void free_plugin(int n);
int plugin_loaded(char *);
void deinit_vis_plugin(HWND hwnd, int n);
void plugin_broadcast(ULONG msg, MPARAM mp1, MPARAM mp2);

void amp_quit(HWND hwnd, HAB haba, HMQ hmq);

int _System pm123_getstring(int index, int subindex, size_t bufsize, char *buf);
void _System pm123_control(int index, void *param);

void no_file_loaded(void);
void update_font(void);


/* Global variables */
/* -----------------*/

/* Main window buttons */
static HWND btn_play;                                      /* Play button    */
static HWND btn_stop;                                      /* Stop button    */
static HWND btn_pause;                                     /* Pause button   */
static HWND btn_rew;                                       /* Rewind button  */
static HWND btn_fwd;                                       /* Forward button */
static HWND btn_power;                                     /* Power button   */
static HWND btn_prev;                                      /* Prev button    */
static HWND btn_next;                                      /* Next button    */
static HWND btn_list;                                      /* List button    */
static HWND btn_load;                                      /* Load button    */
static HWND btn_pl;                                        /* Playlist button*/
static HWND btn_fload;                                     /* Load a file    */

/* Playlist handles */
static HWND playlist;                                      /* Playlist      */
#define CNR_PLAYLIST    47                                 /* Playlist id   */

#define TAGWIN          2012
#define FILEZDLG        2013
#define URLDLG          2014
#define EQUALIZER       2015

#define TRACKDLG        2021
#define LB_TRACKS       2022
#define ST_DRIVE        2023
#define CB_DRIVE        2024
#define PB_REFRESH      2025

/* config pages */
#define NOTEDLG         2010
#define ABOUTNB         2011

#define ID_NULL         42
#define CONFIG_PAGE1    3000
#define CB_PLAYONLOAD   3010
#define CB_AUTOUSEPL    3020
#define CB_AUTOPLAYPL   3030
#define CB_TRASHONSEEK  3040
#define RB_SCROLL_INFINITE 3050
#define RB_SCROLL_ONCE     3060
#define RB_SCROLL_DISABLE  3070
#define RB_DISPLAY_FILENAME 3080
#define RB_DISPLAY_ID3TAG   3090
#define RB_DISPLAY_FILEINFO 3100
#define EF_PROXY_URL        3110
#define EF_HTTP_AUTH        3120
#define SB_BUFFERSIZE       3130
#define CB_FILLBUFFER       3140
#define CB_DOCK             3150
#define EF_DOCK             3151

#define CONFIG_PAGE2    4000
#define LB_VISPLUG      4010
#define PB_VIS_ENABLE   4020
#define PB_VIS_UNLOAD   4030
#define ST_VIS_AUTHOR   4040
#define ST_VIS_DESC     4050
#define PB_VIS_CONFIG   4060
#define PB_VIS_ADD      4070

#define LB_DECPLUG      4080
#define PB_DEC_ENABLE   4090
#define PB_DEC_UNLOAD   4100
#define ST_DEC_AUTHOR   4110
#define ST_DEC_DESC     4120
#define PB_DEC_CONFIG   4130
#define PB_DEC_ADD      4140

#define CONFIG_PAGE3    5000
#define LB_FILPLUG      5010
#define PB_FIL_ENABLE   5020
#define PB_FIL_UNLOAD   5030
#define ST_FIL_AUTHOR   5040
#define ST_FIL_DESC     5050
#define PB_FIL_CONFIG   5060
#define PB_FIL_ADD      5070

#define LB_OUTPLUG      5080
#define PB_OUT_ACTIVATE 5090
#define PB_OUT_UNLOAD   5100
#define ST_OUT_AUTHOR   5110
#define ST_OUT_DESC     5120
#define PB_OUT_CONFIG   5130
#define PB_OUT_ADD      5140

/* specana.c FFT functions */
int _System specana_init(int setnumsamples);
int _System specana_dobands(float bands[]);

/* 123_msg.c */
typedef struct
{
   char *filename;
   char *out_filename;
   char *drive;
   char track;
   HWND hMain;
   char *decoder_needed;

} MSG_PLAY_STRUCT;
void amp_msg(int msg, void *param, void *param2);
void audioSetVolume(int volume);
void StopWait(void);
int TotalTime(void);
int TimePlayed(void);

void equalize_sound(float *gains, BOOL *mute, float preamp, BOOL enabled);


