/*
 * Copyright 1997-2003 Samuel Audet <guardia@step.polymtl.ca>
 *                     Taneli Lepp <rosmo@sektori.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* PM123 Skin Utility */

#define INCL_PM
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "gbm.h"
#include "gbmrect.h"

#include <utilfct.h>

char suberror[256];
char exe[CCHMAXPATH];

#define SKINPAK  -1824
#define SKINPAK2 1824

/* Prototypes */
void dtlf(char *);
int cropimage(char *, char *, int, int, int, int);
int skin_bundle(char *, char *);
int skin_convert(char *);

typedef struct {
  int           resource;
  char          filename[128];
  unsigned long length;
} bundle_hdr;

#define CHECK(x) if (!(x)) { printf("\n\x1b[1mError\x1b[0m: %s\n", suberror); return 0; }
#define DOT printf(".")
#define TSAR(x) putchar(x)

int fexists(char *filename)
{
 struct stat z;
 if (stat(filename, &z) == -1) return 0; else return 1;
}

char *concat(char *st, char *nd)
{
 static char concatbuf[512];
 strcpy(concatbuf, st);
 strcat(concatbuf, nd);
 return(concatbuf);
}

int skin_convert(char *src)
{
 char dir[256], buf[256], buff[SKINPAK2 + 16];
 FILE *skin, *f, *foo, *bar;
 int have_specana = 0;

 strcpy(dir, src);
 if (dir[strlen(dir) - 1] == '\\') dir[strlen(dir) - 1] = '/';
 if (dir[strlen(dir) - 1] != '\\' && dir[strlen(dir) - 1] != '/') strcat(dir, "/");

 sprintf(buf, "%smain.bmp", dir);
 if (!fexists(buf))
  {
   printf("%s: File doesn't exists. The directory doesn't contain a WinAmp skin.\n", buf);
   return 0;
  }
 sprintf(buf, "%scbuttons.bmp", dir);
 if (!fexists(buf))
  {
   printf("%s: File doesn't exists. The directory doesn't contain a WinAmp skin.\n", buf);
   return 0;
  }

 printf("Processing %scbuttons.bmp.", dir);
 sprintf(buf, "%scbuttons.bmp", dir);
 DOT;
 CHECK(cropimage(buf, concat(dir, "prevup.bmp"),  0, 18, 23, 18));
 CHECK(cropimage(buf, concat(dir, "prevdn.bmp"),  0,  0, 23, 18));

 DOT;
 CHECK(cropimage(buf, concat(dir, "playup.bmp"),  23, 18, 23, 18));
 CHECK(cropimage(buf, concat(dir, "playdn.bmp"),  23,  0, 23, 18));

 DOT;
 CHECK(cropimage(buf, concat(dir, "pauseup.bmp"), 45, 18, 23, 18));
 CHECK(cropimage(buf, concat(dir, "pausedn.bmp"), 45,  0, 23, 18));

 DOT;
 CHECK(cropimage(buf, concat(dir, "nextup.bmp"),  91, 18, 23, 18));
 CHECK(cropimage(buf, concat(dir, "nextdn.bmp"),  91,  0, 23, 18));

 DOT;
 CHECK(cropimage(buf, concat(dir, "stopup.bmp"),  68, 18, 23, 18));
 CHECK(cropimage(buf, concat(dir, "stopdn.bmp"),  68,  0, 23, 18));

 DOT;
 CHECK(cropimage(buf, concat(dir, "loadup.bmp"),  114, 20, 22, 16));
 CHECK(cropimage(buf, concat(dir, "loaddn.bmp"),  114,  4, 22, 16));

 printf("\n");

 printf("Processing %stitlebar.bmp.", dir);
 sprintf(buf, "%stitlebar.bmp", dir);

 DOT;
 CHECK(cropimage(buf, concat(dir, "tbd.bmp"),  27, 58, 275, 14));
 CHECK(cropimage(buf, concat(dir, "tb.bmp"),   27, 73, 275, 14));

 DOT;
 CHECK(cropimage(buf, concat(dir, "stb.bmp"),  27, 44, 275, 14));
 CHECK(cropimage(buf, concat(dir, "stbd.bmp"), 27, 30, 275, 14));

 DOT;
 CHECK(cropimage(buf, concat(dir, "tstb.bmp"), 27, 16, 275, 14));
 CHECK(cropimage(buf, concat(dir, "tstbd.bmp"),27,  0, 275, 14));

 DOT;
 CHECK(cropimage(buf, concat(dir, "sprevup.bmp"), 194, 46, 10, 11));
 CHECK(cropimage(buf, concat(dir, "sprevdn.bmp"), 194, 46, 10, 11));

 DOT;
 CHECK(cropimage(buf, concat(dir, "splayup.bmp"), 203, 46, 10, 11));
 CHECK(cropimage(buf, concat(dir, "splaydn.bmp"), 203, 46, 10, 11));

 DOT;
 CHECK(cropimage(buf, concat(dir, "spauseup.bmp"), 213, 46, 10, 11));
 CHECK(cropimage(buf, concat(dir, "spausedn.bmp"), 213, 46, 10, 11));

 DOT;
 CHECK(cropimage(buf, concat(dir, "snextup.bmp"), 232, 46, 10, 11));
 CHECK(cropimage(buf, concat(dir, "snextdn.bmp"), 232, 46, 10, 11));

 DOT;
 CHECK(cropimage(buf, concat(dir, "splup.bmp"), 242, 46, 10, 11));
 CHECK(cropimage(buf, concat(dir, "spldn.bmp"), 242, 46, 10, 11));

 DOT;
 CHECK(cropimage(buf, concat(dir, "pwrup.bmp"), 18, 69, 9, 9));
 CHECK(cropimage(buf, concat(dir, "pwrdn.bmp"), 18, 69, 9, 9));

 printf("\n");

 printf("Processing %stext.bmp: ", dir);
 sprintf(buf, "%stext.bmp", dir);

 TSAR('a');
 CHECK(cropimage(buf, concat(dir, "a.bmp"),   0, 12, 5, 6));
 TSAR('b');
 CHECK(cropimage(buf, concat(dir, "b.bmp"),   5, 12, 5, 6));
 TSAR('c');
 CHECK(cropimage(buf, concat(dir, "c.bmp"),  10, 12, 5, 6));
 TSAR('d');
 CHECK(cropimage(buf, concat(dir, "d.bmp"),  15, 12, 5, 6));
 TSAR('e');
 CHECK(cropimage(buf, concat(dir, "e.bmp"),  20, 12, 5, 6));
 TSAR('f');
 CHECK(cropimage(buf, concat(dir, "f.bmp"),  25, 12, 5, 6));
 TSAR('g');
 CHECK(cropimage(buf, concat(dir, "g.bmp"),  30, 12, 5, 6));
 TSAR('h');
 CHECK(cropimage(buf, concat(dir, "h.bmp"),  35, 12, 5, 6));
 TSAR('i');
 CHECK(cropimage(buf, concat(dir, "i.bmp"),  40, 12, 5, 6));
 TSAR('j');
 CHECK(cropimage(buf, concat(dir, "j.bmp"),  45, 12, 5, 6));
 TSAR('k');
 CHECK(cropimage(buf, concat(dir, "k.bmp"),  50, 12, 5, 6));
 TSAR('l');
 CHECK(cropimage(buf, concat(dir, "l.bmp"),  55, 12, 5, 6));
 TSAR('m');
 CHECK(cropimage(buf, concat(dir, "m.bmp"),  60, 12, 5, 6));
 TSAR('n');
 CHECK(cropimage(buf, concat(dir, "n.bmp"),  65, 12, 5, 6));
 TSAR('o');
 CHECK(cropimage(buf, concat(dir, "o.bmp"),  70, 12, 5, 6));
 TSAR('p');
 CHECK(cropimage(buf, concat(dir, "p.bmp"),  75, 12, 5, 6));
 TSAR('q');
 CHECK(cropimage(buf, concat(dir, "q.bmp"),  80, 12, 5, 6));
 TSAR('r');
 CHECK(cropimage(buf, concat(dir, "r.bmp"),  85, 12, 5, 6));
 TSAR('s');
 CHECK(cropimage(buf, concat(dir, "s.bmp"),  90, 12, 5, 6));
 TSAR('t');
 CHECK(cropimage(buf, concat(dir, "t.bmp"),  95, 12, 5, 6));
 TSAR('u');
 CHECK(cropimage(buf, concat(dir, "u.bmp"), 100, 12, 5, 6));
 TSAR('v');
 CHECK(cropimage(buf, concat(dir, "v.bmp"), 105, 12, 5, 6));
 TSAR('w');
 CHECK(cropimage(buf, concat(dir, "w.bmp"), 110, 12, 5, 6));
 TSAR('x');
 CHECK(cropimage(buf, concat(dir, "x.bmp"), 115, 12, 5, 6));
 TSAR('y');
 CHECK(cropimage(buf, concat(dir, "y.bmp"), 120, 12, 5, 6));
 TSAR('z');
 CHECK(cropimage(buf, concat(dir, "z.bmp"), 125, 12, 5, 6));
 TSAR('_');
 CHECK(cropimage(buf, concat(dir, "_.bmp"),  90,  6, 5, 6));
 TSAR('-');
 CHECK(cropimage(buf, concat(dir, "line.bmp"),  75,  6, 5, 6));
 TSAR('(');
 CHECK(cropimage(buf, concat(dir, "parenl.bmp"),  65,  6, 5, 6));
 TSAR(')');
 CHECK(cropimage(buf, concat(dir, "parenr.bmp"),  70,  6, 5, 6));
 TSAR('&');
 CHECK(cropimage(buf, concat(dir, "amp.bmp"),  125,  6, 5, 6));
 TSAR('.');
 CHECK(cropimage(buf, concat(dir, "dot.bmp"),  135,  6, 5, 6));
 TSAR(':');
 CHECK(cropimage(buf, concat(dir, "colon.bmp"),  60,  6, 5, 6));
 TSAR('0');
 CHECK(cropimage(buf, concat(dir, "0.bmp"),   0,  6, 5, 6));
 TSAR('1');
 CHECK(cropimage(buf, concat(dir, "1.bmp"),   5,  6, 5, 6));
 TSAR('2');
 CHECK(cropimage(buf, concat(dir, "2.bmp"),  10,  6, 5, 6));
 TSAR('3');
 CHECK(cropimage(buf, concat(dir, "3.bmp"),  15,  6, 5, 6));
 TSAR('4');
 CHECK(cropimage(buf, concat(dir, "4.bmp"),  20,  6, 5, 6));
 TSAR('5');
 CHECK(cropimage(buf, concat(dir, "5.bmp"),  25,  6, 5, 6));
 TSAR('6');
 CHECK(cropimage(buf, concat(dir, "6.bmp"),  30,  6, 5, 6));
 TSAR('7');
 CHECK(cropimage(buf, concat(dir, "7.bmp"),  35,  6, 5, 6));
 TSAR('8');
 CHECK(cropimage(buf, concat(dir, "8.bmp"),  40,  6, 5, 6));
 TSAR('9');
 CHECK(cropimage(buf, concat(dir, "9.bmp"),  45,  6, 5, 6));
 TSAR('');
 CHECK(cropimage(buf, concat(dir, ".bmp"),   5,  0, 5, 6));
 TSAR('');
 CHECK(cropimage(buf, concat(dir, ".bmp"),  10,  0, 5, 6));
 TSAR(' ');
 CHECK(cropimage(buf, concat(dir, "space.bmp"), 25,  0, 5, 6));
 TSAR(',');
 CHECK(cropimage(buf, concat(dir, "plat.bmp"),  135,  6, 5, 6));
 TSAR('+');
 CHECK(cropimage(buf, concat(dir, "plus.bmp"),  95,  6, 5, 6));
 TSAR('%');
 CHECK(cropimage(buf, concat(dir, "percent.bmp"),  125,  6, 5, 6));
 TSAR('[');
 CHECK(cropimage(buf, concat(dir, "2parenl.bmp"),  110,  6, 5, 6));
 TSAR(']');
 CHECK(cropimage(buf, concat(dir, "2parenr.bmp"),  115,  6, 5, 6));

 printf("\n");
 printf("Processing %sposbar.bmp.", dir);
 sprintf(buf, "%sposbar.bmp", dir);

 DOT;
 CHECK(cropimage(buf, concat(dir, "slider.bmp"), 248, 0, 29, 10));
 DOT;
 CHECK(cropimage(buf, concat(dir, "shaft.bmp"), 0, 0, 247, 10));

 printf("\n");
 printf("Processing %svolume.bmp.", dir);
 sprintf(buf, "%svolume.bmp", dir);

 DOT;
 CHECK(cropimage(buf, concat(dir, "volbar.bmp"), 0, 15, 68, 13));

 DOT;
 CHECK(cropimage(buf, concat(dir, "volhndl.bmp"), 15, 0, 14, 11));

 printf("\n");
 printf("Processing %sshufrep.bmp:", dir);
 sprintf(buf, "%sshufrep.bmp", dir);

 DOT;
 CHECK(cropimage(buf, concat(dir, "repup.bmp"),  0, 70, 29, 14));
 CHECK(cropimage(buf, concat(dir, "repdn.bmp"),  0, 40, 29, 14));
 DOT;
 CHECK(cropimage(buf, concat(dir, "rndup.bmp"), 29, 70, 46, 14));
 CHECK(cropimage(buf, concat(dir, "rnddn.bmp"), 29, 40, 46, 14));
 DOT;
 CHECK(cropimage(buf, concat(dir, "equp.bmp"),   0, 12, 23, 12));
 CHECK(cropimage(buf, concat(dir, "eqdn.bmp"),   0,  0, 23, 12));
 DOT;
 CHECK(cropimage(buf, concat(dir, "plup.bmp"),  23, 12, 23, 12));
 CHECK(cropimage(buf, concat(dir, "pldn.bmp"),  23,  0, 23, 12));

 printf("\n");
 printf("Processing %smonoster.bmp:", dir);
 sprintf(buf, "%smonoster.bmp", dir);
 CHECK(cropimage(buf, concat(dir, "stereo.bmp"),  0,  12, 28, 12));
 CHECK(cropimage(buf, concat(dir, "mono.bmp"),   28,  12, 28, 12));
 DOT;
 cropimage(buf, concat(dir, "stereo.bmp"),  0,  12, 29, 12);
 cropimage(buf, concat(dir, "mono.bmp"),   29,  12, 29, 12);
 DOT;
 cropimage(buf, concat(dir, "stereo.bmp"),  0,  21, 29, 12);
 cropimage(buf, concat(dir, "mono.bmp"),   29,  21, 29, 12);

 printf("\n");
 printf("Processing %snumbers.bmp:", dir);
 sprintf(buf, "%snumbers.bmp", dir);
 DOT;
 CHECK(cropimage(buf, concat(dir, "t0.bmp"),  0,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t1.bmp"),  9,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t2.bmp"), 18,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t3.bmp"), 27,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t4.bmp"), 36,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t5.bmp"), 45,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t6.bmp"), 54,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t7.bmp"), 63,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t8.bmp"), 72,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "t9.bmp"), 81,  0, 9, 13));
 DOT;
 CHECK(cropimage(buf, concat(dir, "td.bmp"), 90,  0, 1, 1));
 DOT;
 CHECK(cropimage(buf, concat(dir, "tdd.bmp"),90,  0, 1, 1));

 printf("\n");
 printf("Processing %smain.bmp:", dir);
 sprintf(buf, "%smain.bmp", dir);
 DOT;
 CHECK(cropimage(buf, concat(dir, "bg.bmp"), 0, 0, 275, 116));

 printf("\n");

 strcpy(buf, dir);
 strcat(buf, "viscolor.txt");
 f = fopen(buf, "r");
 if (f != NULL)
  {
   strcpy(buf, dir);
   buf[strlen(buf) - 1] = 0;
   strcat(buf, ".dat");
   printf("Creating %s...", buf);
   foo = fopen(buf, "w");
   if (foo == NULL)
    {
     printf("%s: Couldn't open file.\n", buf);
     return 0;
    }
   while (!feof(f))
    {
     fgets(buf, sizeof(buf), f);
     fprintf(foo, "%s", buf);
    }
   fclose(f);
   fclose(foo);
   have_specana = 1;
   printf("\n");
  }

 strcpy(buf, dir);
 buf[strlen(buf) - 1] = 0;
 strcat(buf, ".pak");
 printf("Creating %s\n", buf);

 skin = fopen("pm123.$$$", "w");
 fprintf(skin,"# WinAmp skin conversion bundle template.\n");
 fprintf(skin,"1900,%sbg.bmp\n", dir);
 fprintf(skin,"1630,%svolbar.bmp\n", dir);
 fprintf(skin,"1820,%stb.bmp\n", dir);
 fprintf(skin,"1821,%stbd.bmp\n", dir);
 fprintf(skin,"1905,%sstat.bmp\n", dir);
 fprintf(skin,"1906,%sshaft.bmp\n", dir);
 fprintf(skin,"1300,%splaydn.bmp\n", dir);
 fprintf(skin,"1310,%splayup.bmp\n", dir);
 fprintf(skin,"1302,%spausedn.bmp\n", dir);
 fprintf(skin,"1312,%spauseup.bmp\n", dir);
 fprintf(skin,"1303,%srewdn.bmp\n", dir);
 fprintf(skin,"1313,%srewup.bmp\n", dir);
 fprintf(skin,"1304,%sfwddn.bmp\n", dir);
 fprintf(skin,"1314,%sfwdup.bmp\n", dir);
 fprintf(skin,"1321,%spldn.bmp\n", dir);
 fprintf(skin,"1320,%splup.bmp\n", dir);
 fprintf(skin,"1318,%srndup.bmp\n", dir);
 fprintf(skin,"1308,%srnddn.bmp\n", dir);
 fprintf(skin,"1319,%srepup.bmp\n", dir);
 fprintf(skin,"1309,%srepdn.bmp\n", dir);
 fprintf(skin,"1317,%snextup.bmp\n", dir);
 fprintf(skin,"1307,%snextdn.bmp\n", dir);
 fprintf(skin,"1316,%sprevup.bmp\n", dir);
 fprintf(skin,"1306,%sprevdn.bmp\n", dir);
 fprintf(skin,"1315,%spwrup.bmp\n", dir);
 fprintf(skin,"1305,%spwrdn.bmp\n", dir);
 fprintf(skin,"1323,%sstopup.bmp\n", dir);
 fprintf(skin,"1322,%sstopdn.bmp\n", dir);
 fprintf(skin,"1324,%sloaddn.bmp\n", dir);
 fprintf(skin,"1325,%sloadup.bmp\n", dir);
 fprintf(skin,"1611,%sstb.bmp\n", dir);
 fprintf(skin,"1612,%ststb.bmp\n", dir);
 fprintf(skin,"5500,%ssplaydn.bmp\n", dir);
 fprintf(skin,"5510,%ssplayup.bmp\n", dir);
 fprintf(skin,"5502,%sspausedn.bmp\n", dir);
 fprintf(skin,"5512,%sspauseup.bmp\n", dir);
 fprintf(skin,"5503,%ssrewdn.bmp\n", dir);
 fprintf(skin,"5513,%ssrewup.bmp\n", dir);
 fprintf(skin,"5504,%ssfwddn.bmp\n", dir);
 fprintf(skin,"5514,%ssfwdup.bmp\n", dir);
 fprintf(skin,"5521,%sspldn.bmp\n", dir);
 fprintf(skin,"5520,%ssplup.bmp\n", dir);
 fprintf(skin,"5518,%ssrndup.bmp\n", dir);
 fprintf(skin,"5508,%ssrnddn.bmp\n", dir);
 fprintf(skin,"5519,%ssrepup.bmp\n", dir);
 fprintf(skin,"5509,%ssrepdn.bmp\n", dir);
 fprintf(skin,"5517,%ssnextup.bmp\n", dir);
 fprintf(skin,"5507,%ssnextdn.bmp\n", dir);
 fprintf(skin,"5516,%ssprevup.bmp\n", dir);
 fprintf(skin,"5506,%ssprevdn.bmp\n", dir);
 fprintf(skin,"1400,%sa.bmp\n", dir);
 fprintf(skin,"1401,%sb.bmp\n", dir);
 fprintf(skin,"1402,%sc.bmp\n", dir);
 fprintf(skin,"1403,%sd.bmp\n", dir);
 fprintf(skin,"1404,%se.bmp\n", dir);
 fprintf(skin,"1405,%sf.bmp\n", dir);
 fprintf(skin,"1406,%sg.bmp\n", dir);
 fprintf(skin,"1407,%sh.bmp\n", dir);
 fprintf(skin,"1408,%si.bmp\n", dir);
 fprintf(skin,"1409,%sj.bmp\n", dir);
 fprintf(skin,"1410,%sk.bmp\n", dir);
 fprintf(skin,"1411,%sl.bmp\n", dir);
 fprintf(skin,"1412,%sm.bmp\n", dir);
 fprintf(skin,"1413,%sn.bmp\n", dir);
 fprintf(skin,"1414,%so.bmp\n", dir);
 fprintf(skin,"1415,%sp.bmp\n", dir);
 fprintf(skin,"1416,%sq.bmp\n", dir);
 fprintf(skin,"1417,%sr.bmp\n", dir);
 fprintf(skin,"1418,%ss.bmp\n", dir);
 fprintf(skin,"1419,%st.bmp\n", dir);
 fprintf(skin,"1420,%su.bmp\n", dir);
 fprintf(skin,"1421,%sv.bmp\n", dir);
 fprintf(skin,"1422,%sw.bmp\n", dir);
 fprintf(skin,"1423,%sx.bmp\n", dir);
 fprintf(skin,"1424,%sy.bmp\n", dir);
 fprintf(skin,"1425,%sz.bmp\n", dir);
 fprintf(skin,"1426,%sline.bmp\n", dir);
 fprintf(skin,"1427,%s_.bmp\n", dir);
 fprintf(skin,"1428,%samp.bmp\n", dir);
 fprintf(skin,"1429,%sdot.bmp\n", dir);
 fprintf(skin,"1430,%s0.bmp\n", dir);
 fprintf(skin,"1431,%s1.bmp\n", dir);
 fprintf(skin,"1432,%s2.bmp\n", dir);
 fprintf(skin,"1433,%s3.bmp\n", dir);
 fprintf(skin,"1434,%s4.bmp\n", dir);
 fprintf(skin,"1435,%s5.bmp\n", dir);
 fprintf(skin,"1436,%s6.bmp\n", dir);
 fprintf(skin,"1437,%s7.bmp\n", dir);
 fprintf(skin,"1438,%s8.bmp\n", dir);
 fprintf(skin,"1439,%s9.bmp\n", dir);
 fprintf(skin,"1440,%sspace.bmp\n", dir);
 fprintf(skin,"1441,%sparenl.bmp\n", dir);
 fprintf(skin,"1442,%sparenr.bmp\n", dir);
 fprintf(skin,"1443,%s.bmp\n", dir);
 fprintf(skin,"1444,%s.bmp\n", dir);
 fprintf(skin,"1445,%scolon.bmp\n", dir);
 fprintf(skin,"1446,%splat.bmp\n", dir);
 fprintf(skin,"1447,%splus.bmp\n", dir);
 fprintf(skin,"1448,%spercent.bmp\n", dir);
 fprintf(skin,"1449,%s2parenl.bmp\n", dir);
 fprintf(skin,"1450,%s2parenr.bmp\n", dir);
 fprintf(skin,"1800,%sbnorate.bmp\n", dir);
 fprintf(skin,"1602,%smono.bmp\n", dir);
 fprintf(skin,"1601,%sstereo.bmp\n", dir);
 fprintf(skin,"1608,%snoch.bmp\n", dir);
 fprintf(skin,"1901,%stldark.bmp\n", dir);
 fprintf(skin,"1902,%stl.bmp\n", dir);
 fprintf(skin,"1904,%spleft.bmp\n", dir);
 fprintf(skin,"1903,%spleftd.bmp\n", dir);
 fprintf(skin,"1640,%ss0.bmp\n", dir);
 fprintf(skin,"1641,%ss1.bmp\n", dir);
 fprintf(skin,"1642,%ss2.bmp\n", dir);
 fprintf(skin,"1643,%ss3.bmp\n", dir);
 fprintf(skin,"1644,%ss4.bmp\n", dir);
 fprintf(skin,"1645,%ss5.bmp\n", dir);
 fprintf(skin,"1646,%ss6.bmp\n", dir);
 fprintf(skin,"1647,%ss7.bmp\n", dir);
 fprintf(skin,"1648,%ss8.bmp\n", dir);
 fprintf(skin,"1649,%ss9.bmp\n", dir);
 fprintf(skin,"1650,%ssdd.bmp\n", dir);
 fprintf(skin,"1651,%ss0dark.bmp\n", dir);
 fprintf(skin,"1652,%ssddark.bmp\n", dir);
 fprintf(skin,"1200,%st0.bmp\n", dir);
 fprintf(skin,"1201,%st1.bmp\n", dir);
 fprintf(skin,"1202,%st2.bmp\n", dir);
 fprintf(skin,"1203,%st3.bmp\n", dir);
 fprintf(skin,"1204,%st4.bmp\n", dir);
 fprintf(skin,"1205,%st5.bmp\n", dir);
 fprintf(skin,"1206,%st6.bmp\n", dir);
 fprintf(skin,"1207,%st7.bmp\n", dir);
 fprintf(skin,"1208,%st8.bmp\n", dir);
 fprintf(skin,"1209,%st9.bmp\n", dir);
 fprintf(skin,"1210,%std.bmp\n", dir);
 fprintf(skin,"1211,%stdd.bmp\n", dir);
 fprintf(skin,"1610,%sslider.bmp\n", dir);
 fprintf(skin,"1660,%s0.bmp\n", dir);
 fprintf(skin,"1661,%s1.bmp\n", dir);
 fprintf(skin,"1662,%s2.bmp\n", dir);
 fprintf(skin,"1663,%s3.bmp\n", dir);
 fprintf(skin,"1664,%s4.bmp\n", dir);
 fprintf(skin,"1665,%s5.bmp\n", dir);
 fprintf(skin,"1666,%s6.bmp\n", dir);
 fprintf(skin,"1667,%s7.bmp\n", dir);
 fprintf(skin,"1668,%s8.bmp\n", dir);
 fprintf(skin,"1669,%s9.bmp\n", dir);
 fprintf(skin,"1830,%s0.bmp\n", dir);
 fprintf(skin,"1831,%s1.bmp\n", dir);
 fprintf(skin,"1832,%s2.bmp\n", dir);
 fprintf(skin,"1833,%s3.bmp\n", dir);
 fprintf(skin,"1834,%s4.bmp\n", dir);
 fprintf(skin,"1835,%s5.bmp\n", dir);
 fprintf(skin,"1836,%s6.bmp\n", dir);
 fprintf(skin,"1837,%s7.bmp\n", dir);
 fprintf(skin,"1838,%s8.bmp\n", dir);
 fprintf(skin,"1839,%s9.bmp\n", dir);
 fprintf(skin,"1630,%svolbar.bmp\n", dir);
 fprintf(skin,"1620,%svolhndl.bmp\n", dir);
 strcpy(buf, dir);
 buf[strlen(buf) - 1] = 0;
 strcat(buf, ".dat");

 fclose(skin);

 strcpy(buf, dir);
 buf[strlen(buf) - 1] = 0;
 skin_bundle("pm123.$$$", buf);
 remove("pm123.$$$");

 strcpy(buf, dir);
 buf[strlen(buf) - 1] = 0;

 strcat(buf, ".skn");
 printf("Creating %s\n", buf);
 skin = fopen(buf, "w");
 strcpy(buf, dir);
 buf[strlen(buf) - 1] = 0;
 fprintf(skin, "; Created by PM123 Skin Utility (c) Taneli Lepp <rosmo@sektori.com>\n");
 fprintf(skin, ";\n");
 fprintf(skin, "30,%s.pak\n", buf);
 fprintf(skin, "6,1\n");
 fprintf(skin, "7,1\n");
 fprintf(skin, "8,WinAmp skin\n");
 fprintf(skin, "9,1\n");
 fprintf(skin, "10,1\n");
 fprintf(skin, "20,31\n");
 fprintf(skin, "21,220\n");
 fprintf(skin, "22,16\n");
 fprintf(skin, "23,0\n");
 fprintf(skin, "24,4\n");
 fprintf(skin, "25,0\n");
 fprintf(skin, "26,1\n");
 fprintf(skin, "27,1\n");
 fprintf(skin, "28,1\n");
 fprintf(skin, "29,1\n");
 fprintf(skin, ";\n");
 fprintf(skin, "; 1: Big timer\n");
 fprintf(skin, "1:48,77\n");
 fprintf(skin, "; 2: Main window size (regular)\n");
 fprintf(skin, "2:275,116\n");
 fprintf(skin, "; 3: Play button (regular)\n");
 fprintf(skin, "3:39,9\n");
 fprintf(skin, "; 4: Pause button (regular)\n");
 fprintf(skin, "4:61,9\n");
 fprintf(skin, "; 5: Rewind button (regular)\n");
 fprintf(skin, "5:-1,-1\n");
 fprintf(skin, "; 6: Fast forward button (regular)\n");
 fprintf(skin, "6:-1,-1\n");
 fprintf(skin, "; 7: Playlist button (regular)\n");
 fprintf(skin, "7:241,45\n");
 fprintf(skin, "; 8: Repeat button (regular)\n");
 fprintf(skin, "8:211,12\n");
 fprintf(skin, "; 9: Shuffle button (regular)\n");
 fprintf(skin, "9:164,12\n");
 fprintf(skin, "; 10: Previous button (regular)\n");
 fprintf(skin, "10:16,9\n");
 fprintf(skin, "; 11: Next button (regular)\n");
 fprintf(skin, "11:107,9\n");
 fprintf(skin, "; 12: Power button (regular)\n");
 fprintf(skin, "12:263,103\n");
 fprintf(skin, "; 13: Text display location, regular\n");
 fprintf(skin, "13:111,83\n");
 fprintf(skin, "; 14: Text display location, small\n");
 fprintf(skin, "14:79,4\n");
 fprintf(skin, "; 15: Time left, dark\n");
 fprintf(skin, "15:-1,-1\n");
 fprintf(skin, "; 16: Time left, bright\n");
 fprintf(skin, "16:-1,-1\n");
 fprintf(skin, "; 17: List left, dark\n");
 fprintf(skin, "17:-1,-1\n");
 fprintf(skin, "; 18: List left, bright\n");
 fprintf(skin, "18:-1,-1\n");
 fprintf(skin, "; 19: Time left counter location\n");
 fprintf(skin, "19:-1,-1\n");
 fprintf(skin, "; 20: List left counter location\n");
 fprintf(skin, "20:-1,-1\n");
 fprintf(skin, "; 21: Play mode indicator\n");
 fprintf(skin, "21:-1,-1\n");
 fprintf(skin, "; 22: Led, bright\n");
 fprintf(skin, "22:0,102\n");
 fprintf(skin, "; 23: Led, dark\n");
 fprintf(skin, "23:0,102\n");
 fprintf(skin, "; 24: Slider position\n");
 fprintf(skin, "24:16,35\n");
 fprintf(skin, "; 25: Volume bar position\n");
 fprintf(skin, "25:106,46\n");
 fprintf(skin, "; 26: No channels indicator\n");
 fprintf(skin, "26:-1,-1\n");
 fprintf(skin, "; 27: Mono indicator\n");
 fprintf(skin, "27:211,63\n");
 fprintf(skin, "; 28: Stereo indicator\n");
 fprintf(skin, "28:239,63\n");
 fprintf(skin, "; 29: Bitrate indicator\n");
 fprintf(skin, "29:110,67\n");
 fprintf(skin, "; 30: Main window size (small)\n");
 fprintf(skin, "30:275,14\n");
 fprintf(skin, "; 31: Main window size (tiny)\n");
 fprintf(skin, "31:275,14\n");
 fprintf(skin, "; 33: Play button (small)\n");
 fprintf(skin, "33:175,1\n");
 fprintf(skin, "; 34: Pause button (small)\n");
 fprintf(skin, "34:185,1\n");
 fprintf(skin, "; 35: Rewind button (small)\n");
 fprintf(skin, "35:-1,-1\n");
 fprintf(skin, "; 36: Fast forward button (small)\n");
 fprintf(skin, "36:-1,-1\n");
 fprintf(skin, "; 37: Playlist button (small)\n");
 fprintf(skin, "37:213,1\n");
 fprintf(skin, "; 38: Repeat button (small)\n");
 fprintf(skin, "38:-1,-1\n");
 fprintf(skin, "; 39: Shuffle button (small)\n");
 fprintf(skin, "39:-1,-1\n");
 fprintf(skin, "; 40: Previous button (small)\n");
 fprintf(skin, "40:165,1\n");
 fprintf(skin, "; 41: Next button (small)\n");
 fprintf(skin, "41:203,1\n");
 fprintf(skin, "; 42: Power button (small)\n");
 fprintf(skin, "42:263,2\n");
 fprintf(skin, ";\n");
 fprintf(skin, "; 53: Play button (tiny)\n");
 fprintf(skin, "53:-1,-1\n");
 fprintf(skin, "; 54: Pause button (tiny)\n");
 fprintf(skin, "54:-1,-1\n");
 fprintf(skin, "; 55: Rewind button (tiny)\n");
 fprintf(skin, "55:-1,-1\n");
 fprintf(skin, "; 56: Fast forward button (tiny)\n");
 fprintf(skin, "56:-1,-1\n");
 fprintf(skin, "; 57: Playlist button (tiny)\n");
 fprintf(skin, "57:-1,-1\n");
 fprintf(skin, "; 58: Repeat button (tiny)\n");
 fprintf(skin, "58:-1,-1\n");
 fprintf(skin, "; 59: Shuffle button (tiny)\n");
 fprintf(skin, "59:-1,-1\n");
 fprintf(skin, "; 60: Previous button (tiny)\n");
 fprintf(skin, "60:-1,-1\n");
 fprintf(skin, "; 61: Next button (tiny)\n");
 fprintf(skin, "61:-1,-1\n");
 fprintf(skin, "; 62: Power button (tiny)\n");
 fprintf(skin, "62:263,2\n");
 fprintf(skin, "; 63: Playlist indicator position (index)\n");
 fprintf(skin, "63:-1,-1\n");
 fprintf(skin, "; 64: Playlist indicator position (total)\n");
 fprintf(skin, "64:-1,-1\n");
 fprintf(skin, "; 65: Stop button location (regular)\n");
 fprintf(skin, "65:84,9\n");
 fprintf(skin, "; 66: File load button location (regular)\n");
 fprintf(skin, "66:135,11\n");
 fprintf(skin, "; 67: Slider shaft position, used only if bitmap 1906 is loaded\n");
 fprintf(skin, "67:15,35\n");
 fprintf(skin, "; Plugins\n");
 if (have_specana)
   fprintf(skin,"1=visplug/analyzer.dll,22,56,82,19,%s.dat\n", buf);
  else
   fprintf(skin,"1=visplug/analyzer.dll,22,56,82,19\n");
 fclose(skin);

 printf("Cleaning up...\n");
 remove(concat(dir,"prevup.bmp"));
 remove(concat(dir,"prevdn.bmp"));
 remove(concat(dir,"playup.bmp"));
 remove(concat(dir,"prevdn.bmp"));
 remove(concat(dir,"pauseup.bmp"));
 remove(concat(dir,"pausedn.bmp"));
 remove(concat(dir,"nextup.bmp"));
 remove(concat(dir,"nextdn.bmp"));
 remove(concat(dir,"stopup.bmp"));
 remove(concat(dir,"stopdn.bmp"));
 remove(concat(dir,"loadup.bmp"));
 remove(concat(dir,"loaddn.bmp"));
 remove(concat(dir,"tbd.bmp"));
 remove(concat(dir,"tb.bmp"));
 remove(concat(dir,"stb.bmp"));
 remove(concat(dir,"stbd.bmp"));
 remove(concat(dir,"tstb.bmp"));
 remove(concat(dir,"tstbd.bmp"));
 remove(concat(dir,"sprevup.bmp"));
 remove(concat(dir,"splayup.bmp"));
 remove(concat(dir,"spauseup.bmp"));
 remove(concat(dir,"snextup.bmp"));
 remove(concat(dir,"splup.bmp"));
 remove(concat(dir,"sprevdn.bmp"));
 remove(concat(dir,"splaydn.bmp"));
 remove(concat(dir,"spausedn.bmp"));
 remove(concat(dir,"snextdn.bmp"));
 remove(concat(dir,"spldn.bmp"));
 remove(concat(dir,"volbar.bmp"));
 remove(concat(dir,"volhndl.bmp"));
 remove(concat(dir,"pwrup.bmp"));
 remove(concat(dir,"pwrdn.bmp"));
 remove(concat(dir,"playdn.bmp"));
 remove(concat(dir,"bg.bmp"));
 remove(concat(dir,"a.bmp"));
 remove(concat(dir,"b.bmp"));
 remove(concat(dir,"c.bmp"));
 remove(concat(dir,"d.bmp"));
 remove(concat(dir,"e.bmp"));
 remove(concat(dir,"f.bmp"));
 remove(concat(dir,"g.bmp"));
 remove(concat(dir,"h.bmp"));
 remove(concat(dir,"i.bmp"));
 remove(concat(dir,"j.bmp"));
 remove(concat(dir,"k.bmp"));
 remove(concat(dir,"l.bmp"));
 remove(concat(dir,"m.bmp"));
 remove(concat(dir,"n.bmp"));
 remove(concat(dir,"o.bmp"));
 remove(concat(dir,"p.bmp"));
 remove(concat(dir,"q.bmp"));
 remove(concat(dir,"r.bmp"));
 remove(concat(dir,"s.bmp"));
 remove(concat(dir,"t.bmp"));
 remove(concat(dir,"u.bmp"));
 remove(concat(dir,"v.bmp"));
 remove(concat(dir,"w.bmp"));
 remove(concat(dir,"x.bmp"));
 remove(concat(dir,"y.bmp"));
 remove(concat(dir,"z.bmp"));
 remove(concat(dir,".bmp"));
 remove(concat(dir,".bmp"));
 remove(concat(dir,"_.bmp"));
 remove(concat(dir,"line.bmp"));
 remove(concat(dir,"parenl.bmp"));
 remove(concat(dir,"parenr.bmp"));
 remove(concat(dir,"amp.bmp"));
 remove(concat(dir,"dot.bmp"));
 remove(concat(dir,"colon.bmp"));
 remove(concat(dir,"0.bmp"));
 remove(concat(dir,"1.bmp"));
 remove(concat(dir,"2.bmp"));
 remove(concat(dir,"3.bmp"));
 remove(concat(dir,"4.bmp"));
 remove(concat(dir,"5.bmp"));
 remove(concat(dir,"6.bmp"));
 remove(concat(dir,"7.bmp"));
 remove(concat(dir,"8.bmp"));
 remove(concat(dir,"9.bmp"));
 remove(concat(dir,"space.bmp"));

 remove(concat(dir,"repup.bmp"));
 remove(concat(dir,"repdn.bmp"));
 remove(concat(dir,"rndup.bmp"));
 remove(concat(dir,"rnddn.bmp"));
 remove(concat(dir,"equp.bmp"));
 remove(concat(dir,"eqdn.bmp"));
 remove(concat(dir,"plup.bmp"));
 remove(concat(dir,"pldn.bmp"));

 remove(concat(dir,"slider.bmp"));
 remove(concat(dir,"volbar.bmp"));

 remove(concat(dir,"stereo.bmp"));
 remove(concat(dir,"mono.bmp"));

 remove(concat(dir,"t0.bmp"));
 remove(concat(dir,"t1.bmp"));
 remove(concat(dir,"t2.bmp"));
 remove(concat(dir,"t3.bmp"));
 remove(concat(dir,"t4.bmp"));
 remove(concat(dir,"t5.bmp"));
 remove(concat(dir,"t6.bmp"));
 remove(concat(dir,"t7.bmp"));
 remove(concat(dir,"t8.bmp"));
 remove(concat(dir,"t9.bmp"));
 remove(concat(dir,"td.bmp"));
 remove(concat(dir,"tdd.bmp"));

 remove(concat(dir,"2parenr.bmp"));
 remove(concat(dir,"2parenl.bmp"));
 remove(concat(dir,"percent.bmp"));
 remove(concat(dir,"plus.bmp"));
 remove(concat(dir,"plat.bmp"));

 remove(concat(dir,"shaft.bmp"));

 strcpy(buf, dir);
 buf[strlen(buf) - 1] = 0;
// printf("Finishing %s.pak...\n", buf);

 f = fopen(exe, "rb");
 if (f == NULL)
  {
   printf("%s: Couldn't open file.\n", exe);
   return 0;
  }
 fseek(f, SKINPAK, SEEK_END);
 fread(buff, 1, SKINPAK2, f);
 fclose(f);

 strcpy(buf, dir);
 buf[strlen(buf) - 1] = 0;
 strcat(buf, ".pak");
 f = fopen(buf, "ab");
 if (f == NULL)
  {
   printf("%s: Couldn't open file.\n", buf);
   return 0;
  }
 fwrite(buff, 1, SKINPAK2, f);
 fclose(f);

 return 1;
}

int main(int argc, char *argv[])
{
 getExeName(exe, sizeof(exe));

 printf("\x1b[1mPM123 Skin Utility\x1b[0m (c) 1998-2000 Taneli Lepp <rosmo@sektori.com>\n");
 if (argc < 3)
  {
   printf("Usage: %s {convert|bundle} {parameters}\n\n", argv[0]);
   printf("\x1b[1mconvert\x1b[0m        Converts a WinAmp skin to a PM123 skin\n");
   printf("                 Usage:    convert {skin directory}\n");
   printf("                 Example:  skinutil convert foo\n");
   printf("\x1b[1mbundle\x1b[0m         Bundles skin bitmaps into one file\n");
   printf("                 Usage:    bundle {source.skn} {destination base name}\n");
   printf("                 Example:  skinutil bundle foo.skn foodist\n");
   return 1;
  }
 if (stricmp(argv[1], "bundle") == 0)
  {
   if(skin_bundle(argv[2], argv[3]))
      return 0;
   else
      return 2;
  }
 else if (stricmp(argv[1], "convert") == 0)
  {
   if(skin_convert(argv[2]))
      return 0;
   else
      return 2;
  }

 fprintf(stderr,"\x1b%s\x1b: Unknown command.");
 return 3;
}


void dtlf(char *fixorz) {
  while (fixorz[strlen(fixorz)-1] == 10 || fixorz[strlen(fixorz)]   == 10  ||
         fixorz[strlen(fixorz)-1] == 13 || fixorz[strlen(fixorz)]   == 13)
   {
    if (fixorz[strlen(fixorz)-1] == 10) fixorz[strlen(fixorz) - 1] = '\0';
    if (fixorz[strlen(fixorz)]   == 10) fixorz[strlen(fixorz)]     = '\0';
    if (fixorz[strlen(fixorz)-1] == 13) fixorz[strlen(fixorz) - 1] = '\0';
    if (fixorz[strlen(fixorz)]   == 13) fixorz[strlen(fixorz)]     = '\0';
    }
}

int cropimage(char *src, char *dst, int xx, int yy, int ww, int hh)
{
 char    fn_src[500+1], fn_dst[500+1], *opt_src, *opt_dst;
 int     fd, ft_src, ft_dst, i, stride_src, flag, bytes;
 int     x, y, w, h;
 GBM_ERR rc;
 GBMFT   gbmft;
 GBM gbm;
 GBMRGB  gbmrgb[0x100];
 byte    *data;

 x = xx; y = yy; w = ww; h = hh;

 strcpy(fn_src, src);
 strcpy(fn_dst, dst);

 opt_src = ""; opt_dst = "";

 gbm_init();

 if ( gbm_guess_filetype(fn_src, &ft_src) != GBM_ERR_OK )
  {
   sprintf(suberror, "Cannot guess bitmap file format for %s", fn_src);
   return 0;
  }

 if ( gbm_guess_filetype(fn_dst, &ft_dst) != GBM_ERR_OK )
  {
   sprintf(suberror, "Cannot guess bitmap file format for %s", fn_dst);
   return 0;
  }

 if ( (fd = gbm_io_open(fn_src, O_RDONLY|O_BINARY)) == -1 )
  {
   sprintf(suberror, "Cannot open %s", fn_src);
   return 0;
  }

 if ((rc = gbm_read_header(fn_src, fd, ft_src, &gbm, opt_src)) != GBM_ERR_OK)
  {
   gbm_io_close(fd);
   sprintf(suberror, "Cannot read header of %s: %s", fn_src, gbm_err(rc));
   return 0;
  }

 if ( w     == 0     ) { gbm_io_close(fd); sprintf(suberror, "w = 0"); return 0; }
 if ( h     == 0     ) { gbm_io_close(fd); sprintf(suberror, "h = 0"); return 0; }
 if ( x     >= gbm.w ) { gbm_io_close(fd); sprintf(suberror, "x >= bitmap width"); return 0; }
 if ( y     >= gbm.h ) { gbm_io_close(fd); sprintf(suberror, "y >= bitmap height"); return 0; }
 if ( x + w >  gbm.w ) { gbm_io_close(fd); sprintf(suberror, "x+w > bitmap width"); return 0; }
 if ( y + h >  gbm.h ) { gbm_io_close(fd); sprintf(suberror, "y+h > bitmap height"); return 0; }

 gbm_query_filetype(ft_dst, &gbmft);
 switch ( gbm.bpp )
  {
   case 24:    flag = GBM_FT_W24;  break;
   case 8:     flag = GBM_FT_W8;   break;
   case 4:     flag = GBM_FT_W4;   break;
   case 1:     flag = GBM_FT_W1;   break;
   }

 if ( (gbmft.flags & flag) == 0 )
  {
   gbm_io_close(fd);
   sprintf(suberror, "Output bitmap format %s does not support writing %d bpp data",
       gbmft.short_name, gbm.bpp);
   return 0;
  }

 if ( (rc = gbm_read_palette(fd, ft_src, &gbm, gbmrgb)) != GBM_ERR_OK )
   {
     gbm_io_close(fd);
     sprintf(suberror, "Cannot read palette of %s: %s", fn_src, gbm_err(rc));
     return 0;
   }

 stride_src = ( ((gbm.w * gbm.bpp + 31)/32) * 4 );
 bytes = stride_src * gbm.h;
 if ( (data = malloc((size_t) bytes)) == NULL )
  {
   gbm_io_close(fd);
   sprintf(suberror, "Out of memory allocating %d bytes for bitmap", bytes);
   return 0;
  }

 if ( (rc = gbm_read_data(fd, ft_src, &gbm, data)) != GBM_ERR_OK )
  {
   gbm_io_close(fd);
   sprintf(suberror, "Cannot read bitmap data of %s: %s", fn_src, gbm_err(rc));
   return 0;
  }

 gbm_io_close(fd);

 gbm_subrectangle(&gbm, x, y, w, h, data, data);

 if ( (fd = gbm_io_create(fn_dst, O_WRONLY|O_BINARY)) == -1 )
  {
   sprintf(suberror, "Cannot create %s", fn_dst);
   return 0;
  }

 gbm.w = w;
 gbm.h = h;

 if ( (rc = gbm_write(fn_dst, fd, ft_dst, &gbm, gbmrgb, data, opt_dst)) != GBM_ERR_OK )
  {
   gbm_io_close(fd);
   remove(fn_dst);
   sprintf(suberror, "Cannot write %s: %s", fn_dst, gbm_err(rc));
   return 0;
  }

 gbm_io_close(fd);

 free(data);

 gbm_deinit();

 return 1;
}

int skin_bundle(char *src, char *dst)
{
 FILE *f, *debug, *d, *q, *tak;
 char buf[256], *buf2, buf3[256], *buf4, buf5[256], orig[256];
 int i, images = 0;
 char c;
 unsigned long bwritten = 0;
 bundle_hdr hdr;
 char fbuf[32768];

 f = fopen(src, "r");
 if (f == NULL)
  {
   printf("%s: File not found.\n", src);
   return 0;
  }
 sprintf(buf, "%s.skn", dst);
 if (strcmp(src, buf) == 0)
  {
   printf("Please name the destination skin with a different base name.\n");
   printf("(cannot overwrite %s with %s, you see)\n", src, buf);
   return 0;
  }

 sprintf(buf, "%s.pak", dst);
 d = fopen(buf, "wb");
 if (d == NULL)
  {
   printf("%s: Cannot create file.\n", buf);
   return 0;
  }

 sprintf(buf, "%s.skn", dst);
 tak = fopen(buf, "w");
 if (tak == NULL)
  {
   printf("%s: Cannot create file.\n", buf);
   return 0;
  }
 fprintf(tak, "; Created with PM123 Skin Utility (C) 1998-2000 Taneli Lepp\n");
 fprintf(tak, "; The following line enables the use of the pack file:\n");
 fprintf(tak, "30,%s.pak\n", dst);

 *buf = '\0';

 printf("Bundling...\n");
 /* Background */
 while (!feof(f))
  {
   *buf = '\0';
   while (!*buf || buf[0] == '#' || buf[0] == ';')
    {
     if (*buf)
      fprintf(tak, "%s\n", buf);
     if (feof(f)) break;
     fgets(buf, 256, f);
     if (*buf) dtlf(buf);
     strcpy(orig, buf);
    }

   if (*buf && buf[0] != '#' || buf[0] != ';')
    {
     buf4 = buf;

     while (*buf4 != 0 && *buf4 != ',' && *buf4 != ':' && *buf4 != '=') { buf4++; }

     if (*buf4)
      {
       c = *buf4;
       *buf4 = '\0';
       buf4++;

       /* Bitmap */
       if (*buf && atoi(buf) > 0 && c == ',')
        {
         if (atoi(buf) > 1000)
          {
           q = fopen(buf4, "rb");
           if (q != NULL)
            {
             /* Seek to end */
             fseek(q, 0, SEEK_END);
             hdr.length = ftell(q);
             /* We don't want crap */
             memset(hdr.filename, '\0', 128);
             strcpy(hdr.filename, buf4);
             hdr.resource = atoi(buf);

             /* Seek to beginning */
             fseek(q, 0, SEEK_SET);
             fwrite(&hdr, 1, sizeof(bundle_hdr), d);

             while (!feof(q))
              {
               i = fread(fbuf, 1, sizeof(fbuf), q);
               fwrite(fbuf, 1, i, d);
              }
             fclose(q);
             images++;
             bwritten += hdr.length + sizeof(bundle_hdr);
            }
          } else
            if (atoi(buf) != 30) fprintf(tak, "%s\n", orig);
        } else fprintf(tak, "%s\n", orig);
       }
     } else fprintf(tak, "%s\n", orig);
  }

 fflush(f); fflush(d); fflush(tak);
 fclose(f); fclose(d); fclose(tak);
 printf("%u images, total %u bytes (%u kB).\n", images, bwritten, (bwritten/1024));
 return 1;
}

